"""Test suite module for prance."""

__author__ = 'Jens Finkhaeuser'
__copyright__ = 'Copyright (c) 2016-2018 Jens Finkhaeuser'
__license__ = 'MIT'
__all__ = ()

def _find_imports(*args):
  """
  Helper sorting the named modules into existing and not existing.
  """
  import importlib
  exists = {
    True: [],
    False: [],
  }

  for name in args:
    name = name.replace('-', '_')
    try:
      importlib.import_module(name)
      exists[True].append(name)
    except ImportError:
      exists[False].append(name)

  return exists


def none_of(*args):
  """
  Return true if none of the named modules exist, false otheriwse.
  """
  exists = _find_imports(*args)
  return len(exists[True]) == 0


def _platform(platform, *args):
  """
  Helper for platform()
  """
  ret = False
  for arg in args:
    if arg[0] == '!':
      arg = arg[1:]
      if platform != arg:
        ret = True
        break
    elif platform == arg:
       ret = True
       break

  return ret


def platform(*args):
  """
  Return true if the current platform is in one of the given ones.

  Platforms may be specified with leading '!', in which case they
  a match returns False.
  """
  import sys
  return _platform(sys.platform, *args)



"""Test suite for prance.util.resolver ."""

__author__ = 'Jens Finkhaeuser'
__copyright__ = 'Copyright (c) 2016-2021 Jens Finkhaeuser'
__license__ = 'MIT'
__all__ = ()


import pytest
from unittest.mock import patch

from prance.util import fs
from prance.util import resolver
from prance.util.url import ResolutionError

from . import none_of

def get_specs(fname):
  specs = fs.read_file(fname)

  from prance.util import formats
  specs = formats.parse_spec(specs, fname)

  return specs


def mock_get_petstore(*args, **kwargs):
  from .mock_response import MockResponse, PETSTORE_YAML
  return MockResponse(text = PETSTORE_YAML)


def recursion_limit_handler_none(limit, refstring, recursions):
  return None


@pytest.fixture
def externals_file():
  return get_specs('tests/specs/with_externals.yaml')


@pytest.fixture
def recursive_objs_file():
  return get_specs('tests/specs/recursive_objs.yaml')


@pytest.fixture
def recursive_files_file():
  return get_specs('tests/specs/recursive_files.yaml')


@pytest.fixture
def recursion_limit_file():
  return get_specs('tests/specs/recursion_limit.yaml')


@pytest.fixture
def recursion_limit_files_file():
  return get_specs('tests/specs/recursion_limit_files.yaml')


@pytest.fixture
def missing_file():
  return get_specs('tests/specs/missing_reference.yaml')


def test_resolver_noname(externals_file):
  res = resolver.RefResolver(externals_file)
  # Can't build a fragment URL without reference
  with pytest.raises(ResolutionError):
    res.resolve_references()


@patch('requests.get')
def test_resolver_named(mock_get, externals_file):
  mock_get.side_effect = mock_get_petstore

  import os.path
  from prance.util import fs
  res = resolver.RefResolver(externals_file,
      fs.abspath('tests/specs/with_externals.yaml'))
  res.resolve_references()


@patch('requests.get')
def test_resolver_missing_reference(mock_get, missing_file):
  mock_get.side_effect = mock_get_petstore

  import os.path
  res = resolver.RefResolver(missing_file,
      fs.abspath('tests/specs/missing_reference.yaml'))
  with pytest.raises(ResolutionError) as exc:
    res.resolve_references()

  assert str(exc.value).startswith('Cannot resolve')


@patch('requests.get')
def test_resolver_recursive_objects(mock_get, recursive_objs_file):
  mock_get.side_effect = mock_get_petstore

  # Recursive references to objects are a problem
  import os.path
  res = resolver.RefResolver(recursive_objs_file,
      fs.abspath('tests/specs/recursive_objs.yaml'))
  with pytest.raises(ResolutionError) as exc:
    res.resolve_references()

  assert str(exc.value).startswith('Recursion reached limit')


@patch('requests.get')
def test_resolver_recursive_files(mock_get, recursive_files_file):
  mock_get.side_effect = mock_get_petstore

  # Recursive references to files are not a problem
  import os.path
  res = resolver.RefResolver(recursive_files_file,
      fs.abspath('tests/specs/recursive_files.yaml'))
  res.resolve_references()


def test_recursion_limit_do_not_recurse_raise(recursion_limit_file):
  # Expect the default behaviour to raise.
  import os.path
  res = resolver.RefResolver(recursion_limit_file,
      fs.abspath('tests/specs/recursion_limit.yaml'))
  with pytest.raises(ResolutionError) as exc:
    res.resolve_references()

  assert str(exc.value).startswith('Recursion reached limit of 1')


def test_recursion_limit_do_not_recurse_ignore(recursion_limit_file):
  # If we overload the handler, we should not get an error but should
  # also simply not have the 'next' field - or it should be None
  import os.path
  res = resolver.RefResolver(recursion_limit_file,
      fs.abspath('tests/specs/recursion_limit.yaml'),
      recursion_limit_handler = recursion_limit_handler_none)
  res.resolve_references()

  from prance.util import formats
  contents = formats.serialize_spec(res.specs, 'foo.yaml')

  # The effect of returning None on recursion limit should be that
  # despite having recursion, the outermost reference to
  # definitions/Pet should get resolved.
  assert 'properties' in res.specs['paths']['/pets']['get']['responses']['200']['schema']

  # However, the 'next' field should not be resolved.
  assert res.specs['paths']['/pets']['get']['responses']['200']['schema']['properties']['next']['schema'] is None


def test_recursion_limit_set_limit_ignore(recursion_limit_file):
  # If we overload the handler, and set the recursion limit higher,
  # we should get nested Pet objects a few levels deep.

  import os.path
  res = resolver.RefResolver(recursion_limit_file,
      fs.abspath('tests/specs/recursion_limit.yaml'),
      recursion_limit = 2,
      recursion_limit_handler = recursion_limit_handler_none)
  res.resolve_references()

  from prance.util import formats
  contents = formats.serialize_spec(res.specs, 'foo.yaml')

  # The effect of returning None on recursion limit should be that
  # despite having recursion, the outermost reference to
  # definitions/Pet should get resolved.
  assert 'properties' in res.specs['paths']['/pets']['get']['responses']['200']['schema']

  # However, the 'next' field should be resolved due to the higher recursion limit
  next_field = res.specs['paths']['/pets']['get']['responses']['200']['schema']['properties']['next']['schema']
  assert next_field is not None

  # But the 'next' field of the 'next' field should not be resolved.
  assert next_field['properties']['next']['schema'] is None


def test_recursion_limit_do_not_recurse_raise_files(recursion_limit_files_file):
  # Expect the default behaviour to raise.
  import os.path
  res = resolver.RefResolver(recursion_limit_files_file,
      fs.abspath('tests/specs/recursion_limit_files.yaml'))
  with pytest.raises(ResolutionError) as exc:
    res.resolve_references()

  assert str(exc.value).startswith('Recursion reached limit of 1')


def test_recursion_limit_do_not_recurse_ignore_files(recursion_limit_files_file):
  # If we overload the handler, we should not get an error but should
  # also simply not have the 'next' field - or it should be None
  import os.path
  res = resolver.RefResolver(recursion_limit_files_file,
      fs.abspath('tests/specs/recursion_limit_files.yaml'),
      recursion_limit_handler = recursion_limit_handler_none)
  res.resolve_references()

  from prance.util import formats
  contents = formats.serialize_spec(res.specs, 'foo.yaml')

  # The effect of returning None on recursion limit should be that
  # despite having recursion, the outermost reference to
  # definitions/Pet should get resolved.
  assert 'properties' in res.specs['paths']['/pets']['get']['responses']['200']['schema']

  # However, the 'next' field should not be resolved.
  assert res.specs['paths']['/pets']['get']['responses']['200']['schema']['properties']['next']['schema'] is None


def test_recursion_limit_set_limit_ignore_files(recursion_limit_files_file):
  # If we overload the handler, and set the recursion limit higher,
  # we should get nested Pet objects a few levels deep.

  import os.path
  res = resolver.RefResolver(recursion_limit_files_file,
      fs.abspath('tests/specs/recursion_limit_files.yaml'),
      recursion_limit = 2,
      recursion_limit_handler = recursion_limit_handler_none)
  res.resolve_references()

  from prance.util import formats
  contents = formats.serialize_spec(res.specs, 'foo.yaml')

  # The effect of returning None on recursion limit should be that
  # despite having recursion, the outermost reference to
  # definitions/Pet should get resolved.
  assert 'properties' in res.specs['paths']['/pets']['get']['responses']['200']['schema']

  # However, the 'next' field should be resolved due to the higher recursion limit
  next_field = res.specs['paths']['/pets']['get']['responses']['200']['schema']['properties']['next']['schema']
  assert next_field is not None

  # But the 'next' field of the 'next' field should not be resolved.
  assert next_field['properties']['next']['schema'] is None

@patch('requests.get')
def test_issue_22_empty_path(mock_get, externals_file):
  mock_get.side_effect = mock_get_petstore

  # The raw externals file must have unresolved data
  assert len(externals_file['paths']['/pets/{petId}']['get']['parameters']) == 1
  param = externals_file['paths']['/pets/{petId}']['get']['parameters'][0]

  assert 'overwritten' in param
  assert '$ref' in param

  import os.path
  from prance.util import fs
  res = resolver.RefResolver(externals_file,
      fs.abspath('tests/specs/with_externals.yaml'))
  res.resolve_references()

  # The tests should resolve the reference, but the reference overwrites
  # all else.
  assert len(res.specs['paths']['/pets/{petId}']['get']['parameters']) == 1
  param = res.specs['paths']['/pets/{petId}']['get']['parameters'][0]

  # Dereferenced keys must exist here
  assert 'type' in param
  assert 'description' in param
  assert 'required' in param
  assert 'in' in param
  assert 'name' in param

  # Previously defined keys must not
  assert 'overwritten' not in param
  assert '$ref' not in param

def test_issue_38_tilde_one():
  specs = get_specs('tests/specs/issue_38_a.yaml')
  res = resolver.RefResolver(specs,
      fs.abspath('tests/specs/issue_38_a.yaml'))
  res.resolve_references()

  path = res.specs['paths']['/api/v2/vms']
  assert 'get' in path
  assert 'operationId' in path['get']
  assert 'description' in path['get']


@patch('requests.get')
def test_issue_23_partial_resolution_all(mock_get):
  mock_get.side_effect = mock_get_petstore

  specs = get_specs('tests/specs/with_externals.yaml')
  res = resolver.RefResolver(specs,
      fs.abspath('tests/specs/with_externals.yaml'))
  res.resolve_references()

  # By default, all externals need to be resolved.
  from prance.util.path import path_get

  val = path_get(res.specs, ('paths', '/pets', 'get', 'responses', '200', 'schema'))
  assert '$ref' not in val

  val = path_get(res.specs, ('paths', '/pets', 'get', 'responses', 'default', 'schema'))
  assert '$ref' not in val

  val = path_get(res.specs, ('paths', '/pets', 'post', 'responses', 'default', 'schema'))
  assert '$ref' not in val

  val = path_get(res.specs, ('paths', '/pets/{petId}', 'get', 'parameters', 0))
  assert '$ref' not in val

  val = path_get(res.specs, ('paths', '/pets/{petId}', 'get', 'responses', '200', 'schema'))
  assert '$ref' not in val

  val = path_get(res.specs, ('paths', '/pets/{petId}', 'get', 'responses', 'default', 'schema'))
  assert '$ref' not in val


def test_issue_23_partial_resolution_internal():
  specs = get_specs('tests/specs/with_externals.yaml')
  res = resolver.RefResolver(specs,
      fs.abspath('tests/specs/with_externals.yaml'),
      resolve_types = resolver.RESOLVE_INTERNAL
      )
  res.resolve_references()

  # By default, all externals need to be resolved.
  from prance.util.path import path_get

  val = path_get(res.specs, ('paths', '/pets', 'get', 'responses', '200', 'schema'))
  assert '$ref' in val

  val = path_get(res.specs, ('paths', '/pets', 'get', 'responses', 'default', 'schema'))
  assert '$ref' in val

  val = path_get(res.specs, ('paths', '/pets', 'post', 'responses', 'default', 'schema'))
  assert '$ref' in val

  val = path_get(res.specs, ('paths', '/pets/{petId}', 'get', 'parameters', 0))
  assert '$ref' in val

  val = path_get(res.specs, ('paths', '/pets/{petId}', 'get', 'responses', '200', 'schema'))
  assert '$ref' in val

  val = path_get(res.specs, ('paths', '/pets/{petId}', 'get', 'responses', 'default', 'schema'))
  assert '$ref' not in val


def test_issue_23_partial_resolution_files():
  specs = get_specs('tests/specs/with_externals.yaml')
  res = resolver.RefResolver(specs,
      fs.abspath('tests/specs/with_externals.yaml'),
      resolve_types = resolver.RESOLVE_FILES
      )
  res.resolve_references()

  # By default, all externals need to be resolved.
  from prance.util.path import path_get

  val = path_get(res.specs, ('paths', '/pets', 'get', 'responses', '200', 'schema'))
  assert '$ref' not in val

  val = path_get(res.specs, ('paths', '/pets', 'get', 'responses', 'default', 'schema'))
  assert '$ref' not in val

  val = path_get(res.specs, ('paths', '/pets', 'post', 'responses', 'default', 'schema'))
  assert '$ref' not in val

  val = path_get(res.specs, ('paths', '/pets/{petId}', 'get', 'parameters', 0))
  assert '$ref' not in val

  val = path_get(res.specs, ('paths', '/pets/{petId}', 'get', 'responses', '200', 'schema'))
  assert '$ref' in val

  val = path_get(res.specs, ('paths', '/pets/{petId}', 'get', 'responses', 'default', 'schema'))
  assert '$ref' in val


@pytest.mark.skipif(none_of('openapi-spec-validator'), reason='Missing backends')
def test_issue_65_partial_resolution_files():
  specs = '''openapi: "3.0.0"
info:
  title: ''
  version: '1.0.0'
paths: {}
components:
    schemas:
        SampleArray:
            type: array
            items:
              $ref: '#/components/schemas/ItemType'

        ItemType:
          type: integer
'''
  from prance.util import formats
  specs = formats.parse_spec(specs, 'issue_65.yaml')

  res = resolver.RefResolver(specs,
      fs.abspath('issue_65.yaml'),
      resolve_types = resolver.RESOLVE_FILES
      )
  res.resolve_references()

  from prance.util.path import path_get
  val = path_get(res.specs, ('components', 'schemas', 'SampleArray', 'items'))
  assert '$ref' in val


@patch('requests.get')
def test_issue_23_partial_resolution_http(mock_get):
  mock_get.side_effect = mock_get_petstore

  specs = get_specs('tests/specs/with_externals.yaml')
  res = resolver.RefResolver(specs,
      fs.abspath('tests/specs/with_externals.yaml'),
      resolve_types = resolver.RESOLVE_HTTP
      )
  res.resolve_references()

  # By default, all externals need to be resolved.
  from prance.util.path import path_get

  val = path_get(res.specs, ('paths', '/pets', 'get', 'responses', '200', 'schema'))
  assert '$ref' in val

  val = path_get(res.specs, ('paths', '/pets', 'get', 'responses', 'default', 'schema'))
  assert '$ref' in val

  val = path_get(res.specs, ('paths', '/pets', 'post', 'responses', 'default', 'schema'))
  assert '$ref' in val

  val = path_get(res.specs, ('paths', '/pets/{petId}', 'get', 'parameters', 0))
  assert '$ref' in val

  val = path_get(res.specs, ('paths', '/pets/{petId}', 'get', 'responses', '200', 'schema'))
  assert '$ref' not in val

  val = path_get(res.specs, ('paths', '/pets/{petId}', 'get', 'responses', 'default', 'schema'))
  assert '$ref' in val


def test_issue_23_partial_resolution_invalid_scheme():
  specs = {'$ref': 'foo://cannot-do-anything'}
  res = resolver.RefResolver(specs,
      fs.abspath('tests/specs/with_externals.yaml'))

  with pytest.raises(ValueError):
    res.resolve_references()

def test_issue_69_urlparse_error():
  # XXX depending on python version, the error may not actually come from
  #     parsing the URL, but from trying to load a local file that does
  #     not exist. See test_issue_72_* for a test case that specifically
  #     tries for a nonexistent file url.
  specs = {'$ref': "file://a\u2100b/bad/netloc"}
  res = resolver.RefResolver(specs,
      fs.abspath('tests/specs/with_externals.yaml'))

  with pytest.raises(ResolutionError) as exinfo:
    res.resolve_references()

  assert 'bad/netloc' in str(exinfo.value)


def test_issue_72_nonexistent_file_error():
  specs = {'$ref': "file://does/not/exist"}
  res = resolver.RefResolver(specs,
      fs.abspath('tests/specs/with_externals.yaml'))

  with pytest.raises(ResolutionError) as exinfo:
    res.resolve_references()

  assert 'not/exist' in str(exinfo.value)



@pytest.mark.skipif(none_of('openapi-spec-validator'), reason='Missing backends')
def test_issue_78_resolve_internal_bug():
  specs = ''
  with open('tests/specs/issue_78/openapi.json') as fh:
    specs = fh.read()

  from prance.util import formats
  specs = formats.parse_spec(specs, 'openapi.json')

  res = resolver.RefResolver(specs,
      fs.abspath('openapi.json'),
      resolve_types = resolver.RESOLVE_FILES
      )
  res.resolve_references()

  from prance.util.path import path_get
  val = path_get(res.specs, ('paths', '/endpoint', 'post', 'requestBody', 'content'))

  # Reference to file is resolved
  assert 'application/json' in val
  # Internal reference within file is NOT resolved
  assert '$ref' in val['application/json']



@pytest.mark.skipif(none_of('openapi-spec-validator'), reason='Missing backends')
def test_issue_77_translate_external():
  specs = ''
  with open('tests/specs/issue_78/openapi.json') as fh:
    specs = fh.read()

  from prance.util import formats
  specs = formats.parse_spec(specs, 'openapi.json')

  res = resolver.RefResolver(specs,
      fs.abspath('openapi.json'),
      resolve_types = resolver.RESOLVE_FILES,
      resolve_method= resolver.TRANSLATE_EXTERNAL
      )
  res.resolve_references()

  from prance.util.path import path_get
  val = path_get(res.specs, ('components', 'schemas', '_schemas.json_Body'))

  # Reference to file is translated in components/schemas
  assert 'content' in val
  assert 'application/json' in val['content']

  # Reference url is updated
  val = path_get(res.specs, ('paths', '/endpoint', 'post', 'requestBody', '$ref'))
  assert val == '#/components/schemas/_schemas.json_Body'



@pytest.mark.skipif(none_of('openapi-spec-validator'), reason='Missing backends')
def test_issue_77_translate_external_refs_internal():
  specs = ''
  with open('tests/specs/issue_78/openapi.json') as fh:
    specs = fh.read()

  from prance.util import formats
  specs = formats.parse_spec(specs, 'openapi.json')

  res = resolver.RefResolver(specs,
      fs.abspath('openapi.json'),
      resolve_types = resolver.RESOLVE_FILES | resolver.RESOLVE_INTERNAL,
      resolve_method= resolver.TRANSLATE_EXTERNAL
      )
  res.resolve_references()

  from prance.util.path import path_get
  val = path_get(res.specs, ('components', 'schemas', '_schemas.json_Body'))

  # Reference to file is translated in components/schemas
  assert 'content' in val
  assert 'application/json' in val['content']

  # Internal Reference links updated
  assert '#/components/schemas/_schemas.json_Something' == val['content']['application/json']['$ref']

  # Internal references is copied to componnents/schemas seperately
  val = path_get(res.specs, ('components', 'schemas', '_schemas.json_Something'))
  assert 'type' in val

  # File reference url is updated as well
  val = path_get(res.specs, ('paths', '/endpoint', 'post', 'requestBody', '$ref'))
  assert val == '#/components/schemas/_schemas.json_Body'


@pytest.mark.skipif(none_of('openapi-spec-validator'), reason='Missing backends')
def test_issue_77_internal_refs_unresolved():
  specs = ''
  with open('tests/specs/issue_78/openapi.json') as fh:
    specs = fh.read()

  from prance.util import formats
  specs = formats.parse_spec(specs, 'openapi.json')

  res = resolver.RefResolver(specs,
      fs.abspath('openapi.json'),
      resolve_types = resolver.RESOLVE_FILES,
      resolve_method= resolver.TRANSLATE_EXTERNAL
      )
  res.resolve_references()

  from prance.util.path import path_get
  val = path_get(res.specs, ('components', 'schemas'))

  # File reference resolved
  assert '_schemas.json_Body' in val

  # Internal file reference not resolved
  assert '_schemas.json_Something' not in val

@pytest.mark.xfail(reason="https://github.com/RonnyPfannschmidt/prance/issues/87")
def test_issue_205_swagger_resolution_failure():
  specs = ''
  with open('tests/specs/kubernetes_api_docs.json') as fh:
    specs = fh.read()

  from prance.util import formats
  specs = formats.parse_spec(specs, 'kubernetes_api_docs.json')

  res = resolver.RefResolver(specs,
    fs.abspath('kubernetes_api_docs.json'),
    resolve_types = resolver.RESOLVE_FILES,
    resolve_method= resolver.TRANSLATE_EXTERNAL
  )
  # test will raise an exception
  res.resolve_references()



"""Test suite for prance.util.iterators ."""

__author__ = 'Jens Finkhaeuser'
__copyright__ = 'Copyright (c) 2016-2021 Jens Finkhaeuser'
__license__ = 'MIT'
__all__ = ()


from prance.util import iterators

def test_item_iterator():
  tester = {
    'foo': 42,
    'bar': {
      'some': 'dict',
    },
    'baz': [
      { 1: 2 },
      { 3: 4 },
    ]
  }

  frozen = tuple(iterators.item_iterator(tester))
  assert len(frozen) == 9, "Iterator did not yield the correct number of items!"

  # Test for a few paths
  assert ((), tester) in frozen
  assert (('bar', 'some'), 'dict') in frozen
  assert (('baz', 0, 1), 2) in frozen
  assert (('baz', 1), { 3: 4 }) in frozen


def test_item_iterator_empty():
  frozen = tuple(iterators.item_iterator({}))
  assert len(frozen) == 1, "Must return at least the item itself!"
  assert ((), {}) in frozen

  frozen = tuple(iterators.item_iterator([]))
  assert len(frozen) == 1, "Must return at least the item itself!"
  assert ((), []) in frozen


def test_reference_iterator_empty_dict():
  tester = {}
  frozen = tuple(iterators.reference_iterator(tester))
  assert len(frozen) == 0, 'Found items when it should not have!'


def test_reference_iterator_dict_without_references():
  tester = {
    'foo': 42,
    'bar': 'baz',
    'baz': {
      'quux': {
        'fnord': False,
      },
    },
  }
  frozen = tuple(iterators.reference_iterator(tester))
  assert len(frozen) == 0, 'Found items when it should not have!'


def test_reference_iterator_dict_with_references():
  tester = {
    'foo': 42,
    'bar': 'baz',
    '$ref': 'root',
    'baz': {
      '$ref': 'branch',
      'quux': {
        'fnord': False,
        '$ref': 'leaf',
      },
    },
  }
  frozen = tuple(iterators.reference_iterator(tester))
  assert len(frozen) == 3, 'Found a different item count than expected!'

  # We have three references with their paths here, but we don't know which
  # order the tuple has them in. Let's go through them all:
  expectations = {
    0: 'root',
    1: 'branch',
    2: 'leaf',
  }
  for key, value, path in iterators.reference_iterator(tester):
    assert value == expectations[len(path)]



"""Test suite for prance.util.url ."""

__author__ = 'Jens Finkhaeuser'
__copyright__ = 'Copyright (c) 2016-2021 Jens Finkhaeuser'
__license__ = 'MIT'
__all__ = ()

import sys

import pytest

from prance.util import url

from . import platform


def test_absurl_http():
  test = 'http://foo.bar/asdf/#lala/quux'
  res = url.absurl(test)
  assert res.geturl() == test


def test_absurl_http_fragment():
  base = 'http://foo.bar/asdf/#lala/quux'
  test = '#another'
  res = url.absurl(test, base)
  assert res.scheme == 'http'
  assert res.netloc == 'foo.bar'
  assert res.path == '/asdf/'
  assert res.fragment == 'another'


@pytest.mark.skipif(platform('win32'), reason = 'Skip on win32')
def test_absurl_file_posix():
  base = 'file:///etc/passwd'
  test = 'group'
  expect = 'file:///etc/group'
  res = url.absurl(test, base)
  assert res.geturl() == expect


@pytest.mark.skipif(platform('!win32'), reason = 'Skip on !win32')
def test_absurl_file_win32():
  base = 'file:///c:/windows/notepad.exe'
  test = "regedit.exe"
  expect = 'file:///c:/windows/regedit.exe'
  res = url.absurl(test, base)
  assert res.geturl() == expect


@pytest.mark.skipif(platform('win32'), reason = 'Skip on win32')
def test_absurl_absfile_posix():
  test = 'file:///etc/passwd'
  res = url.absurl(test)
  assert res.geturl() == test


@pytest.mark.skipif(platform('!win32'), reason = 'Skip on !win32')
def test_absurl_absfile_win32():
  test = 'file:///c:/windows/notepad.exe'
  res = url.absurl(test)
  assert res.geturl() == test


def test_absurl_fragment():
  base = 'file:///etc/passwd'
  test = '#frag'
  with pytest.raises(url.ResolutionError):
    url.absurl(test)

  res = url.absurl(test, base)
  assert res.geturl() == 'file:///etc/passwd#frag'


def test_absurl_relfile():
  base = 'http://foo.bar'
  test = 'relative.file'
  with pytest.raises(url.ResolutionError):
    url.absurl(test)
  with pytest.raises(url.ResolutionError):
    url.absurl(test, base)


@pytest.mark.skipif(platform('win32'), reason = 'Skip on win32')
def test_absurl_paths_posix():
  base = '/etc/passwd'
  test = 'group'
  expect = 'file:///etc/group'

  with pytest.raises(url.ResolutionError):
    url.absurl(test)

  res = url.absurl(test, base)
  assert res.geturl() == expect


@pytest.mark.skipif(platform('!win32'), reason = 'Skip on !win32')
def test_absurl_paths_win32():
  base = 'c:\\windows\\notepad.exe'
  test = "regedit.exe"
  expect = 'file:///c:/windows/regedit.exe'

  with pytest.raises(url.ResolutionError):
    url.absurl(test)

  res = url.absurl(test, base)
  assert res.geturl() == expect


def test_urlresource():
  parsed = url.absurl('http://foo.bar/asdf?some=query#myfrag')
  res = url.urlresource(parsed)
  assert res == 'http://foo.bar/asdf'


def test_split_url_reference():
  base = url.absurl('http://foo.bar/')

  # Relative reference
  parsed, path = url.split_url_reference(base, '#foo/bar')
  assert parsed.netloc == 'foo.bar'
  assert len(path) == 2
  assert path[0] == 'foo'
  assert path[1] == 'bar'

  # Leading slashes are stripped
  parsed, path = url.split_url_reference(base, '#///foo/bar')
  assert parsed.netloc == 'foo.bar'
  assert len(path) == 2
  assert path[0] == 'foo'
  assert path[1] == 'bar'

  # Absolute reference
  parsed, path = url.split_url_reference(base, 'http://somewhere/#foo/bar')
  assert parsed.netloc == 'somewhere'
  assert len(path) == 2
  assert path[0] == 'foo'
  assert path[1] == 'bar'

  # Reference with escaped parts
  parsed, path = url.split_url_reference(base, 'http://somewhere/#foo/bar~1baz~1quux~0foo')
  assert parsed.netloc == 'somewhere'
  assert len(path) == 2
  assert path[0] == 'foo'
  assert path[1] == 'bar/baz/quux~foo'


def test_fetch_url_file():
  from prance.util import fs
  content = url.fetch_url(url.absurl(fs.abspath('tests/specs/with_externals.yaml')))
  assert content['swagger'] == '2.0'


def test_fetch_url_cached():
  from prance.util import fs
  cache = {}

  content1 = url.fetch_url(url.absurl(fs.abspath('tests/specs/with_externals.yaml')), cache)
  assert content1['swagger'] == '2.0'

  content2 = url.fetch_url(url.absurl(fs.abspath('tests/specs/with_externals.yaml')), cache)
  assert content2['swagger'] == '2.0'

  # Dicts are mutable, therefore we can't compare IDs. But individual
  # string fields should not be copied, because we shallow copy.
  assert id(content1['swagger']) == id(content2['swagger'])


def test_fetch_url_text_cached():
  from prance.util import fs
  cache = {}

  content1, _ = url.fetch_url_text(url.absurl(fs.abspath('tests/specs/with_externals.yaml')), cache)
  content2, _ = url.fetch_url_text(url.absurl(fs.abspath('tests/specs/with_externals.yaml')), cache)

  # Strings are immutable, therefore IDs should be identical
  assert id(content1) == id(content2)


@pytest.mark.requires_network()
def test_fetch_url_http():
  exturl = 'http://finkhaeuser.de/projects/prance/petstore.yaml'\
    '#/definitions/Pet'
  content = url.fetch_url(url.absurl(exturl))
  assert content['swagger'] == '2.0'


def test_fetch_url_python():
  exturl = 'python://tests/specs/petstore.yaml'
  content = url.fetch_url(url.absurl(exturl))
  assert content['swagger'] == '2.0'



"""This submodule contains file format related utility code for Prance."""

__author__ = 'Jens Finkhaeuser'
__copyright__ = 'Copyright (c) 2016-2021 Jens Finkhaeuser'
__license__ = 'MIT'
__all__ = ()


class ParseError(ValueError):
  pass  # pragma: nocover


def __format_preferences(filename, content_type):  # noqa: N802
  """
  Detect the format based on file name and content type.

  Each parameter may be None, so a heuristic can be used in the end.

  :return: A tuple of format strings, in the optimal order to try.
  :rtype: tuple
  """
  # Select the correct format.
  # 1) If there is neither file name nor content type, use a heuristic.
  # 2) If there is a file name but no content type, use the file extension.
  # 3) If there is no file name, but a content type, use the content type.
  # 4) If both are present, prefer the content type.
  # 5) use a heuristic either way to catch bad content types, file names,
  #    etc. The selection process above is just the most likely match!
  best = None

  if filename and not content_type:
    from os.path import splitext
    _, ext = splitext(filename)

    for extensions in __EXT_TO_FORMAT.keys():
      if ext in extensions:
        best = __EXT_TO_FORMAT[extensions]

  elif content_type:
    # Split off the first part of the content type; for us, that's enough.
    content_type = content_type.split(';')[0]

    for ctypes in __MIME_TO_FORMAT.keys():
      if content_type in ctypes:
        best = __MIME_TO_FORMAT[ctypes]

  # If we have no best format yet, we need to use a heuristic. This is tricky;
  # Swagger is largely YAML-based, but JSON is used for remote references. In
  # the end, JSON is probably more likely to match.
  if not best:
    best = 'JSON'

  # Now assemble an ordered list of formats to return, with the best format
  # first.
  formats = list(__EXT_TO_FORMAT.values())
  formats.remove(best)
  formats.insert(0, best)

  return tuple(formats)


# Basic parse functions
def __parse_yaml(spec_str):  # noqa: N802
  import yaml, yaml.parser
  try:
    return yaml.safe_load(str(spec_str))
  except yaml.parser.ParserError as err:
    raise ParseError(str(err))


def __parse_json(spec_str):  # noqa: N802
  import json
  try:
    return json.loads(str(spec_str))
  except ValueError as err:
    raise ParseError(str(err))


# Basic serialization functions
def __serialize_yaml(specs):  # noqa: N802
  # I don't know what the PyYAML authors smoked, but in order for it to parse
  # unicode specs and not mess encoding up, you need to allow_unicode and
  # explicitly set the encoding to None.
  import yaml
  utf = yaml.dump(specs,
                  allow_unicode = True,
                  encoding = None,
                  default_flow_style = False)

  return str(utf)


def __serialize_json(specs):  # noqa: N802
  # The default encoding is utf-8, no need to specify it. But we need to switch
  # off ensure_ascii, otherwise we do not get a unicode string back.
  import json
  utf = json.dumps(specs, ensure_ascii = False, indent = 2)

  return str(utf)


# Map file name extensions to parse/serialize functions
__EXT_TO_FORMAT = {
  ('.yaml', '.yml'): 'YAML',
  ('.json', '.js'): 'JSON',
}

__MIME_TO_FORMAT = {
  ('application/json', 'application/javascript'): 'JSON',
  ('application/x-yaml', 'text/yaml'): 'YAML',
}


__FORMAT_TO_PARSER = {
  'YAML': __parse_yaml,
  'JSON': __parse_json,
}

__FORMAT_TO_SERIALIZER = {
  'YAML': __serialize_yaml,
  'JSON': __serialize_json,
}


def format_info(format_name):
  """
  Return content type and extension for a supported format.

  Valid formats are `YAML` or `JSON`.

  :param str format_name: The name of the format.
  :return: The preferred content type and file name extension, or
      (None, None) if the format name is not supported.
  :rtype: tuple
  """
  format_name = format_name.upper()

  content_type = None
  for content_types, name in __MIME_TO_FORMAT.items():
    if name == format_name:
      content_type = content_types[0]

  extension = None
  for extensions, name in __EXT_TO_FORMAT.items():
    if name == format_name:
       extension = extensions[0]

  return content_type, extension


def parse_spec_details(spec_str, filename = None, **kwargs):
  """
  Return a parsed dict of the given spec string.

  Also returned are the detected mime type and file name extension.

  The default format is assumed to be JSON, but if you provide a filename,
  its extension is used to determine whether YAML or JSON should be
  parsed.

  :param str spec_str: The specifications as string.
  :param str filename: [optional] Filename to determine the format from.
  :param str content_type: [optional] Content type to determine the format
      from.
  :return: The specifications, mime type, and extension.
  :rtype: tuple
  :raises ParseError: when parsing fails.
  """
  # Fetch optional content type & determine formats
  content_type = kwargs.get('content_type', None)
  formats = __format_preferences(filename, content_type)

  # Try parsing each format in order
  for f in formats:
    parser = __FORMAT_TO_PARSER[f]
    try:
      result = parser(spec_str)
      ctype, ext = format_info(f)
      return result, ctype, ext
    except ParseError:
      pass

  # All failed!
  raise ParseError('Could not detect format of spec string!')


def parse_spec(spec_str, filename = None, **kwargs):
  """
  Return a parsed dict of the given spec string.

  The function exists for legacy reasons and just wraps parse_spec_details,
  returning only the parsed specs.

  :param str spec_str: The specifications as string.
  :param str filename: [optional] Filename to determine the format from.
  :param str content_type: [optional] Content type to determine the format
      from.
  :return: The specifications.
  :rtype: dict
  :raises ParseError: when parsing fails.
  """
  result, ctype, ext = parse_spec_details(spec_str, filename, **kwargs)
  return result


def serialize_spec(specs, filename = None, **kwargs):
  """
  Return a serialized version of the given spec.

  The default format is assumed to be JSON, but if you provide a filename,
  its extension is used to determine whether YAML or JSON should be
  parsed.

  :param dict specs: The specifications as dict.
  :param str filename: [optional] Filename to determine the format from.
  :param str content_type: [optional] Content type to determine the format
      from.
  :return: The serialized specifications.
  :rtype: str
  """
  # Fetch optional content type & determine formats
  content_type = kwargs.get('content_type', None)
  formats = __format_preferences(filename, content_type)

  # Instead of trying to parse various formats, we only serialize to the first
  # one in the list - nothing else makes much sense.
  serializer = __FORMAT_TO_SERIALIZER[formats[0]]
  return serializer(specs)



"""Test OpenAPI specs examples."""

__author__ = 'Jens Finkhaeuser'
__copyright__ = 'Copyright (c) 2018 Jens Finkhaeuser'
__license__ = 'MIT'
__all__ = ()

import pytest

from prance import ResolvingParser
from prance import ValidationError
from prance.util.fs import FileNotFoundError

def make_name(path, parser, backend, version, file_format, entry):
  import os.path
  basename = os.path.splitext(entry)[0]
  basename = basename.replace('-', '_')
  version = version.replace('.', '')
  backend = backend.replace('-', '_')

  name = '_'.join(['test', parser, backend, version, file_format, basename])
  return name

# Generate test cases at import.
# One case per combination of:
# - parser (base, resolving)
# - validation backend (flex, swagger-spec-validator) (skip ssv if importing it fails)
# - spec version
# - file format
# - file
# That gives >50 test cases

import os, os.path
base = os.path.join('tests', 'OpenAPI-Specification', 'examples')

def iter_entries(parser, backend, version, file_format, path):
  if version == 'v3.0' and backend != 'openapi-spec-validator':
    return

  for entry in os.listdir(path):
    full = os.path.join(path, entry)
    testcase_name = None
    if os.path.isfile(full):
      testcase_name = make_name(full, parser, backend, version, file_format, entry)
    elif os.path.isdir(full):
      if parser == 'BaseParser':
        continue  # skip separate files for the BaseParser
      full = os.path.join(full, 'spec', 'swagger.%s' % (file_format))
      if os.path.isfile(full):
        testcase_name = make_name(full, parser, backend, version, file_format, entry)

    if testcase_name:
      dirname = os.path.dirname(full)
      dirname = dirname.replace('\\', '\\\\')
      from prance.util import url
      absurl = url.absurl(os.path.abspath(full)).geturl()
      code = """
@pytest.mark.xfail()
def %s():
  import os
  cur = os.getcwd()

  os.chdir('%s')

  from prance import %s
  try:
    parser = %s('%s', backend = '%s')
  finally:
    os.chdir(cur)
""" % (testcase_name, dirname, parser, parser, absurl, backend)
      print(code)
      exec(code, globals())

for parser in ('BaseParser', 'ResolvingParser'):
  from prance.util import validation_backends
  for backend in validation_backends():
    for version in os.listdir(base):
      version_dir = os.path.join(base, version)
      for file_format in os.listdir(version_dir):
        format_dir = os.path.join(version_dir, file_format)

        if not os.path.isdir(format_dir):  # Assume YAML
          iter_entries(parser, backend, version, 'yaml', version_dir)
        else:
          for entry in os.listdir(format_dir):
            iter_entries(parser, backend, version, file_format, format_dir)



"""
Functionality for converting from Swagger/OpenAPI 2.0 to OpenAPI 3.0.0.

The functions use https://mermade.org.uk/ APIs for conversion.
"""

__author__ = 'Jens Finkhaeuser'
__copyright__ = 'Copyright (c) 2018 Jens Finkhaeuser'
__license__ = 'MIT'
__all__ = ()


class ConversionError(ValueError):
  pass  # pragma: nocover


def convert_str(spec_str, filename = None, **kwargs):
  """
  Convert the serialized spec.

  We parse the spec first to ensure there is no parse error, then
  send it off to the API for conversion.

  :param str spec_str: The specifications as string.
  :param str filename: [optional] Filename to determine the format from.
  :param str content_type: [optional] Content type to determine the format
      from.
  :return: The converted spec and content type.
  :rtype: tuple
  :raises ParseError: when parsing fails.
  :raises ConversionError: when conversion fails.
  """
  # Parse, but discard the results. The function raises parse error.
  from .util.formats import parse_spec_details
  spec, content_type, extension = parse_spec_details(spec_str, filename,
      **kwargs)

  # Ok, parsing went fine, so let's convert.
  data = {
    'source': spec_str,
  }
  if filename is not None:
    data['filename'] = filename
  else:
    data['filename'] = f'openapi{extension}'

  headers = {
    'accept': f'{content_type}; charset=utf-8'
  }

  # Convert via API
  import requests
  r = requests.post('https://mermade.org.uk/api/v1/convert', data = data,
      headers = headers)
  if not r.ok:  # pragma: nocover
    raise ConversionError('Could not convert spec: %d %s' % (
        r.status_code, r.reason))

  return r.text, '{}; {}'.format(r.headers['content-type'], r.apparent_encoding)


def convert_url(url, cache = {}):
  """
  Fetch a URL, and try to convert it to OpenAPI 3.x.y.

  :param str url: The URL to fetch.
  :return: The converted spec and content type.
  :rtype: tuple
  :raises ParseError: when parsing fails.
  :raises ConversionError: when conversion fails.
  """
  # Fetch URL contents
  from .util.url import fetch_url_text
  content, content_type = fetch_url_text(url, cache)

  # Try converting
  return convert_str(content, None, content_type = content_type)


def convert_spec(parser_or_spec, parser_klass = None, *args, **kwargs):
  """
  Convert an already parsed spec to OpenAPI 3.x.y.

  Returns a new parser instance with the parsed specs, if possible.

  The first parameter is either a parsed OpenAPI 2.0 spec, or a parser
  instance, i.e. something derived from prance.BaseParser. If a parser,
  the returned parser's options are taken from this source parser.

  If the first parameter is a parsed spec, you must specify the class
  of parser to instanciate. You can specify other options as key word
  arguments. See the parser klass for details.

  Any key word arguments specified here also override options from a
  source parser.

  This parametrization may seem a little convoluted. What it does, though,
  is allow maximum flexibility. You can create parsed (but unvalidated)
  OpenAPI 3.0 specs even if you only have backends that support vresion 2.0.
  You can pass the source parser, and the lazy flag, and that's it. If your
  version 2.0 specs were valid, there's a good chance your converted 3.0
  specs are also valid.

  :param mixed parser_or_spec: A dict (spec) or an instance of BaseParser
  :param type parser_klass: [optional] A parser class to instanciate for
    the result.
  :return: A parser instance.
  :rtype: BaseParser or derived.
  """
  # Figure out exact configuration to use
  klass = None
  options = None
  spec = None

  from . import BaseParser
  if isinstance(parser_or_spec, BaseParser):
    # We have a parser instance
    klass = parser_klass or type(parser_or_spec)
    options = parser_or_spec.options.copy()
    options.update(kwargs)
    spec = parser_or_spec.specification
  else:
    # We must assume a specification
    klass = parser_klass or BaseParser
    options = kwargs.copy()
    spec = parser_or_spec

  # print('Class', klass)
  # print('Options', options)
  # print('Spec', spec)

  # We have to serialize the specs in order to convert them. Let's use
  # YAML.
  from .util import formats
  serialized = formats.serialize_spec(spec, content_type = 'text/yaml')

  # Convert serialized
  converted, ctype = convert_str(serialized, content_type = 'text/yaml')

  # Create parser with options
  result = klass(spec_string = converted, **options)
  return result



"""This submodule contains code for fetching/parsing URLs."""

__author__ = 'Jens Finkhaeuser'
__copyright__ = 'Copyright (c) 2016-2018 Jens Finkhaeuser'
__license__ = 'MIT'
__all__ = ()


from urllib import parse


class ResolutionError(LookupError):
  pass


def urlresource(url):
  """
  Return the resource part of a parsed URL.

  The resource part is defined as the part without query, parameters or
  fragment. Just the scheme, netloc and path remains.

  :param tuple url: A parsed URL
  :return: The resource part of the URL
  :rtype: str
  """
  res_list = list(url)[0:3] + [None, None, None]
  return parse.ParseResult(*res_list).geturl()


def absurl(url, relative_to = None):
  """
  Turn relative file URLs into absolute file URLs.

  This is necessary, because while JSON pointers do not allow relative file
  URLs, Swagger/OpenAPI explicitly does. We need to make relative paths
  absolute before passing them off to jsonschema for verification.

  Non-file URLs are left untouched. URLs without scheme are assumed to be file
  URLs.

  :param str/tuple url: The input URL.
  :param str/tuple relative_to: [optional] The URL to which the input URL is
      relative.
  :return: The output URL, parsed into components.
  :rtype: tuple
  """
  # Parse input URL, if necessary
  parsed = url
  if not isinstance(parsed, tuple):
    from .fs import is_pathname_valid
    if is_pathname_valid(url):
      from . import fs
      url = fs.to_posix(url)
    try:
      parsed = parse.urlparse(url)
    except Exception as ex:
      from .exceptions import raise_from
      raise_from(ResolutionError, ex, f'Unable to parse url: {url}')

  # Any non-file scheme we just return immediately.
  if parsed.scheme not in (None, '', 'file'):
    return parsed

  # Parse up the reference URL
  reference = relative_to
  if reference and not isinstance(reference, tuple):
    from .fs import is_pathname_valid
    if is_pathname_valid(reference):
      from . import fs
      reference = fs.to_posix(reference)
    reference = parse.urlparse(reference)

  # If the input URL has no path, we assume only its fragment matters.
  # That is, we'll have to set the fragment of the reference URL to that
  # of the input URL, and return the result.
  import os.path
  from .fs import from_posix, abspath
  result_list = None
  if not parsed.path:
    if not reference or not reference.path:
      raise ResolutionError('Cannot build an absolute file URL from a fragment'
          ' without a reference with path!')
    result_list = list(reference)
    result_list[5] = parsed.fragment
  elif os.path.isabs(from_posix(parsed.path)):
    # We have an absolute path, so we can ignore the reference entirely!
    result_list = list(parsed)
    result_list[0] = 'file'  # in case it was empty
  else:
    # If we have a relative path, we require a reference.
    if not reference:
      raise ResolutionError('Cannot build an absolute file URL from a relative'
          ' path without a reference!')
    if reference.scheme not in (None, '', 'file'):
      raise ResolutionError('Cannot build an absolute file URL with a non-file'
          ' reference!')

    result_list = list(parsed)
    result_list[0] = 'file'  # in case it was empty
    result_list[2] = abspath(from_posix(parsed.path),
            from_posix(reference.path))

  # Reassemble the result and return it
  result = parse.ParseResult(*result_list)
  return result


def split_url_reference(base_url, reference):
  """
  Return a normalized, parsed URL and object path.

  The reference string is a JSON reference, i.e. a URL with a fragment that
  contains an object path into the referenced resource.

  The base URL is used as a reference point for relative references.

  :param mixed base_url: A parsed URL.
  :param str reference: A JSON reference string.
  :return: The parsed absolute URL of the reference and the object path.
  """
  # Parse URL
  parsed_url = absurl(reference, base_url)

  # Grab object path
  obj_path = parsed_url.fragment.split('/')
  while len(obj_path) and not obj_path[0]:
    obj_path = obj_path[1:]

  # Normalize the object path by substituting ~1 and ~0 respectively.
  def _normalize(path):
    path = path.replace('~1', '/')
    path = path.replace('~0', '~')
    return path
  obj_path = [_normalize(p) for p in obj_path]

  return parsed_url, obj_path


def fetch_url_text(url, cache = {}, encoding = None):
  """
  Fetch the URL.

  If the URL is a file URL, the format used for parsing depends on the file
  extension. Otherwise, YAML is assumed.

  The URL may also use the `python` scheme. In this scheme, the netloc part
  refers to an importable python package, and the path part to a path relative
  to the package path, e.g. `python://some_package/path/to/file.yaml`.

  :param tuple url: The url, parsed as returned by `absurl` above.
  :param Mapping cache: An optional cache. If the URL can be found in the
    cache, return the cache contents.
  :param str encoding: Provide an encoding for local URLs to override
    encoding detection, if desired. Defaults to None.
  :return: The resource text of the URL, and the content type.
  :rtype: tuple
  """
  url_key = 'text_' + urlresource(url)
  entry = cache.get(url_key, None)
  if entry is not None:
    return entry

  # Fetch contents according to scheme. We assume requests can handle all the
  # non-file schemes, or throw otherwise.
  content = None
  content_type = None
  if url.scheme in (None, '', 'file'):
    from .fs import read_file, from_posix
    try:
      content = read_file(from_posix(url.path), encoding)
    except FileNotFoundError as ex:
      from .exceptions import raise_from
      raise_from(ResolutionError, ex, f'File not found: {url.path}')
  elif url.scheme == 'python':
    # Resolve package path
    package = url.netloc
    path = url.path
    if path[0] == '/':
      path = path[1:]

    import pkg_resources
    path = pkg_resources.resource_filename(package, path)

    from .fs import read_file, from_posix
    content = read_file(from_posix(path), encoding)
  else:
    import requests
    response = requests.get(url.geturl())
    if not response.ok:  # pragma: nocover
      raise ResolutionError('Cannot fetch URL "%s": %d %s' % (
        url.geturl(), response.status_code, response.reason))
    content_type = response.headers.get('content-type', 'text/plain')
    content = response.text

  cache[url_key] = (content, content_type)
  return content, content_type


def fetch_url(url, cache = {}, encoding = None, strict = True):
  """
  Fetch the URL and parse the contents.

  Same as fetch_url_text(), but also parses the content and only
  returns the parse results.

  :param tuple url: The url, parsed as returned by `absurl` above.
  :param Mapping cache: An optional cache. If the URL can be found in the
    cache, return the cache contents.
  :param str encoding: Provide an encoding for local URLs to override
    encoding detection, if desired. Defaults to None.
  :return: The parsed file.
  :rtype: dict
  """
  # Return from cache, if parsed result is already present.
  url_key = (urlresource(url), strict)
  entry = cache.get(url_key, None)
  if entry is not None:
    return entry.copy()

  # Fetch URL text
  content, content_type = fetch_url_text(url, cache, encoding = encoding)

  # Parse the result
  from .formats import parse_spec
  result = parse_spec(content, url.path, content_type = content_type)

  # Perform some sanitization in lenient mode.
  if not strict:
    from . import stringify_keys
    result = stringify_keys(result)

  # Cache and return result
  cache[url_key] = result
  return result.copy()



"""Utility code for tests."""

__author__ = 'Jens Finkhaeuser'
__copyright__ = 'Copyright (c) 2021 Jens Finkhaeuser'
__license__ = 'MIT'
__all__ = ()

import pytest

class MockResponse:
  """
  Minimally imitate a response object from the requests library.

  Implements all that's required for prance.util.url
  """

  def __init__(self, *args, **kwargs):
    self.ok = kwargs.get('is_ok', True)
    self.headers = kwargs.get('headers', { 'content-type': 'text/plain' })
    self.text = kwargs.get('text', '')

# The petstore.yaml file served at
# http://finkhaeuser.de/projects/prance/petstore.yaml
PETSTORE_YAML = '''swagger: "2.0"
info:
  version: 1.0.0
  title: Swagger Petstore
  license:
    name: MIT
host: petstore.swagger.io
basePath: /v1
schemes:
  - http
consumes:
  - application/json
produces:
  - application/json
paths:
  /pets:
    get:
      summary: List all pets
      operationId: listPets
      tags:
        - pets
      parameters:
        - name: limit
          in: query
          description: How many items to return at one time (max 100)
          required: false
          type: integer
          format: int32
      responses:
        "200":
          description: An paged array of pets
          headers:
            x-next:
              type: string
              description: A link to the next page of responses
          schema:
            $ref: '#/definitions/Pets'
        default:
          description: unexpected error
          schema:
            $ref: '#/definitions/Error'
    post:
      summary: Create a pet
      operationId: createPets
      tags:
        - pets
      responses:
        "201":
          description: Null response
        default:
          description: unexpected error
          schema:
            $ref: '#/definitions/Error'
  /pets/{petId}:
    get:
      summary: Info for a specific pet
      operationId: showPetById
      tags:
        - pets
      parameters:
        - name: petId
          in: path
          required: true
          description: The id of the pet to retrieve
          type: string
      responses:
        "200":
          description: Expected response to a valid request
          schema:
            $ref: '#/definitions/Pets'
        default:
          description: unexpected error
          schema:
            $ref: '#/definitions/Error'
definitions:
  Pet:
    required:
      - id
      - name
    properties:
      id:
        type: integer
        format: int64
      name:
        type: string
      tag:
        type: string
  Pets:
    type: array
    items:
      $ref: '#/definitions/Pet'
  Error:
    required:
      - code
      - message
    properties:
      code:
        type: integer
        format: int32
      message:
        type: string
'''



"""Test suite focusing on validation backend features."""

__author__ = 'Jens Finkhaeuser'
__copyright__ = 'Copyright (c) 2018 Jens Finkhaeuser'
__license__ = 'MIT'
__all__ = ()

import pytest

from prance import BaseParser
from prance import ValidationError
from prance.util import validation_backends

from . import none_of

def test_bad_backend():
  with pytest.raises(ValueError):
    BaseParser('tests/specs/petstore.yaml', backend = 'does_not_exist')


@pytest.mark.skipif(none_of('flex'), reason='Missing dependencies: flex')
def test_flex_issue_5_integer_keys():
  # Must succeed with default (flex) parser; note the parser does not stringify the response code
  parser = BaseParser('tests/specs/issue_5.yaml', backend = 'flex')
  assert 200 in parser.specification['paths']['/test']['post']['responses']


@pytest.mark.skipif(none_of('flex'), reason='Missing dependencies: flex')
def test_flex_validate_success():
  parser = BaseParser('tests/specs/petstore.yaml', backend = 'flex')


@pytest.mark.skipif(none_of('flex'), reason='Missing dependencies: flex')
def test_flex_validate_failure():
  with pytest.raises(ValidationError):
    parser = BaseParser('tests/specs/missing_reference.yaml', backend = 'flex')


@pytest.mark.skipif(none_of('swagger-spec-validator'), reason='Missing dependencies: swagger-spec-validator')
def test_swagger_spec_validator_issue_5_integer_keys():
  # Must fail in implicit strict mode.
  with pytest.raises(ValidationError):
    BaseParser('tests/specs/issue_5.yaml', backend = 'swagger-spec-validator')

  # Must fail in explicit strict mode.
  with pytest.raises(ValidationError):
    BaseParser('tests/specs/issue_5.yaml', backend = 'swagger-spec-validator', strict = True)

  # Must succeed in non-strict/lenient mode
  parser = BaseParser('tests/specs/issue_5.yaml', backend = 'swagger-spec-validator', strict = False)
  assert '200' in parser.specification['paths']['/test']['post']['responses']


@pytest.mark.skipif(none_of('swagger-spec-validator'), reason='Missing dependencies: swagger-spec-validator')
def test_swagger_spec_validator_validate_success():
  parser = BaseParser('tests/specs/petstore.yaml', backend = 'swagger-spec-validator')


@pytest.mark.skipif(none_of('swagger-spec-validator'), reason='Missing dependencies: swagger-spec-validator')
def test_swagger_spec_validator_validate_failure():
  with pytest.raises(ValidationError):
    parser = BaseParser('tests/specs/missing_reference.yaml', backend = 'swagger-spec-validator')


@pytest.mark.skipif(none_of('openapi-spec-validator'), reason='Missing dependencies: openapi-spec-validator')
def test_openapi_spec_validator_issue_5_integer_keys():
  # Must fail in implicit strict mode.
  with pytest.raises(ValidationError):
    BaseParser('tests/specs/issue_5.yaml', backend = 'openapi-spec-validator')

  # Must fail in explicit strict mode.
  with pytest.raises(ValidationError):
    BaseParser('tests/specs/issue_5.yaml', backend = 'openapi-spec-validator', strict = True)

  # Must succeed in non-strict/lenient mode
  parser = BaseParser('tests/specs/issue_5.yaml', backend = 'openapi-spec-validator', strict = False)
  assert '200' in parser.specification['paths']['/test']['post']['responses']


@pytest.mark.skipif(none_of('openapi-spec-validator'), reason='Missing dependencies: openapi-spec-validator')
def test_openapi_spec_validator_issue_36_error_reporting():
  with pytest.raises(ValidationError, match = r'Strict mode enabled'):
    BaseParser('tests/specs/issue_36.yaml')


@pytest.mark.skipif(none_of('openapi-spec-validator'), reason='Missing dependencies: openapi-spec-validator')
def test_openapi_spec_validator_validate_success():
  parser = BaseParser('tests/specs/petstore.yaml', backend = 'openapi-spec-validator')


@pytest.mark.skipif(none_of('openapi-spec-validator'), reason='Missing dependencies: openapi-spec-validator')
def test_openapi_spec_validator_validate_failure():
  with pytest.raises(ValidationError):
    parser = BaseParser('tests/specs/missing_reference.yaml', backend = 'openapi-spec-validator')


@pytest.mark.skipif(none_of('openapi-spec-validator'), reason='Missing dependencies: openapi-spec-validator')
def test_openapi_spec_validator_issue_20_spec_version_handling():
  # The spec is OpenAPI 3, but broken. Need to set 'strict' to False to stringify keys
  with pytest.raises(ValidationError):
    parser = BaseParser('tests/specs/issue_20.yaml', backend = 'openapi-spec-validator', strict = False)

  # Lazy parsing should let us validate what's happening
  parser = BaseParser('tests/specs/issue_20.yaml', backend = 'openapi-spec-validator', strict = False, lazy = True)
  assert not parser.valid
  assert parser.version_parsed == ()

  with pytest.raises(ValidationError):
    parser.parse()

  # After parsing, the specs are not valid, but the correct version is
  # detected.
  assert not parser.valid
  assert parser.version_parsed == (3, 0, 0)



"""This submodule contains specialty iterators over specs."""

__author__ = 'Jens Finkhaeuser'
__copyright__ = 'Copyright (c) 2016-2018 Jens Finkhaeuser'
__license__ = 'MIT'
__all__ = ()


def item_iterator(value, path = ()):
  """
  Return item iterator over the a nested dict- or list-like object.

  Returns each item value as the second item to unpack, and a tuple path to the
  item as the first value - in that, it behaves much like viewitems(). For list
  like values, the path is made up of numeric indices.

  Given a spec such as this::

    spec = {
      'foo': 42,
      'bar': {
        'some': 'dict',
      },
      'baz': [
        { 1: 2 },
        { 3: 4 },
      ]
    }

  Here, (parts of) the yielded values would be:

    ======== =============
    item     path
    ======== =============
    [...]    ('baz',)
    { 1: 2 } ('baz', 0)
    2        ('baz', 0, 1)
    ======== =============

  :param dict/list value: The specifications to iterate over.
  :return: An iterator over all items in the value.
  :rtype: iterator
  """
  # Yield the top-level object, always
  yield path, value

  from collections.abc import Mapping, Sequence

  # For dict and list like objects, we also need to yield each item
  # recursively.
  if isinstance(value, Mapping):
    for key, item in value.items():
      yield from item_iterator(item, path + (key,))
  elif isinstance(value, Sequence) and not isinstance(value,
         str):
    for idx, item in enumerate(value):
      yield from item_iterator(item, path + (idx,))


def reference_iterator(specs, path = ()):
  """
  Iterate through the given specs, returning only references.

  The iterator returns three values:
    - The key, mimicking the behaviour of other iterators, although
      it will always equal '$ref'
    - The value
    - The path to the item. This is a tuple of all the item's ancestors,
      in sequence, so that you can reasonably easily find the containing
      item. It does not include the final '$ref' key.

  :param dict specs: The specifications to iterate over.
  :return: An iterator over all references in the specs.
  :rtype: iterator
  """
  # We need to iterate through the nested specification dict, so let's
  # start with an appropriate iterator. We can immediately optimize it by
  # only returning '$ref' items.
  for item_path, item in item_iterator(specs, path):
    if len(item_path) <= 0:
      continue
    key = item_path[-1]
    if key == '$ref':
      yield key, item, item_path[:-1]



"""Test suite for prance.BaseParser ."""

__author__ = 'Jens Finkhaeuser'
__copyright__ = 'Copyright (c) 2016-2021 Jens Finkhaeuser'
__license__ = 'MIT'
__all__ = ()

import pytest

from prance import BaseParser
from prance import ValidationError
from prance.util.fs import FileNotFoundError

from . import none_of


@pytest.fixture
def petstore_parser():
  return BaseParser('tests/specs/petstore.yaml')


@pytest.fixture
def petstore_parser_from_string():
  yaml = None
  with open('tests/specs/petstore.yaml', 'rb') as f:
    x = f.read()
    yaml = x.decode('utf8')
  return BaseParser(spec_string = yaml)


def test_load_fail():
  from prance.util.url import ResolutionError
  with pytest.raises(ResolutionError):
    BaseParser('tests/specs/missing.yaml')


def test_parse_fail():
  with pytest.raises(ValidationError):
    BaseParser(spec_string = """---
invalid 'a'sda YAML""")


def test_version_fail():
  with pytest.raises(ValidationError):
    BaseParser(spec_string = """---
openapi: 4.0.0""")


def test_filename_or_spec():
  with pytest.raises(AssertionError):
    BaseParser('', '')


@pytest.mark.skipif(none_of('openapi-spec-validator', 'swagger-spec-validator', 'flex'), reason='Backends missing.')
def test_load_and_parse_valid(petstore_parser):
  assert petstore_parser.specification, 'No specs loaded!'


@pytest.mark.skipif(none_of('openapi-spec-validator', 'swagger-spec-validator', 'flex'), reason='Backends missing.')
def test_load_and_parse_lazy():
  parser = BaseParser('tests/specs/petstore.yaml', lazy = True)
  assert parser.specification is None, 'Should not have specs yet!'

  parser.parse()
  assert parser.specification, 'No specs loaded!'


@pytest.mark.skipif(none_of('openapi-spec-validator', 'swagger-spec-validator', 'flex'), reason='Backends missing.')
def test_yaml_valid(petstore_parser):
  assert petstore_parser.yaml(), 'Did not get YAML representation of specs!'


@pytest.mark.skipif(none_of('openapi-spec-validator', 'swagger-spec-validator', 'flex'), reason='Backends missing.')
def test_json_valid(petstore_parser):
  assert petstore_parser.json(), 'Did not get JSON representation of specs!'


@pytest.mark.skipif(none_of('openapi-spec-validator', 'swagger-spec-validator', 'flex'), reason='Backends missing.')
def test_cache_specs_mixin(petstore_parser):
  # In order to test the caching, we need to first use either the YAML or the
  # JSON mixin. Let's use YAML, because it's more swagger-ish
  yaml = petstore_parser.yaml()
  assert yaml, 'Did not get YAML representation of specs!'

  # Caching should mean that if the specifications do not change, then neither
  # does the YAML repersentation.
  assert yaml == petstore_parser.yaml(), 'YAML representation changed!'

  # In fact, the objects shouldn't even change.
  assert id(yaml) == id(petstore_parser.yaml()), 'YAML did not change but '\
      'got regenerated!'

  # However, when the specs change, then so must the YAML representation.
  petstore_parser.specification['foo'] = 'bar'
  assert yaml != petstore_parser.yaml(), 'YAML representation did not change!'


@pytest.mark.skipif(none_of('openapi-spec-validator', 'swagger-spec-validator', 'flex'), reason='Backends missing.')
def test_relative_urls_from_string(petstore_parser_from_string):
  # This must succeed
  assert petstore_parser_from_string.yaml(), 'Did not get YAML representation of specs!'



"""
Prance implements parsers for Swagger/OpenAPI 2.0 and 3.0.0 API specs.

See https://openapis.org/ for details on the specification.

Included is a BaseParser that reads and validates swagger specs, and a
ResolvingParser that additionally resolves any $ref references.
"""

__author__ = 'Jens Finkhaeuser'
__copyright__ = 'Copyright (c) 2016-2021 Jens Finkhaeuser'
__license__ = 'MIT'
__all__ = ('util', 'mixins', 'cli', 'convert')

try:
  from prance._version import version as __version__
except ImportError:
  # todo: better gussing
  __version__ = '0.20.0+unknown'


# Define our own error class
class ValidationError(Exception):
  pass


from . import mixins

# Placeholder for when no URL is specified for the main spec file
import sys
if sys.platform == 'win32':  # pragma: nocover
  # Placeholder must be absolute
  _PLACEHOLDER_URL = 'file:///c:/__placeholder_url__.yaml'
else:
  _PLACEHOLDER_URL = 'file:///__placeholder_url__.yaml'


class BaseParser(mixins.YAMLMixin, mixins.JSONMixin):
  """
  The BaseParser loads, parses and validates OpenAPI 2.0 and 3.0.0 specs.

  Uses :py:class:`YAMLMixin` and :py:class:`JSONMixin` for additional
  functionality.
  """

  BACKENDS = {
    'flex': ((2,), '_validate_flex'),
    'swagger-spec-validator': ((2,), '_validate_swagger_spec_validator'),
    'openapi-spec-validator': ((2, 3), '_validate_openapi_spec_validator'),
  }

  SPEC_VERSION_2_PREFIX = 'Swagger/OpenAPI'
  SPEC_VERSION_3_PREFIX = 'OpenAPI'

  def __init__(self, url = None, spec_string = None, lazy = False, **kwargs):
    """
    Load, parse and validate specs.

    You can either provide a URL or a spec string, but not both.

    :param str url: The URL of the file to load. URLs missing a scheme are
      assumed to be file URLs.
    :param str spec_string: The specifications to parse.
    :param bool lazy: If true, do not load or parse anything. Instead wait for
      the parse function to be invoked.
    :param str backend: [optional] one of 'flex', 'swagger-spec-validator' or
      'openapi-spec-validator'.
      Determines the validation backend to use. Defaults to the first installed
      backend in the ordered list obtained from util.validation_backends().
    :param bool strict: [optional] Applies only to the 'swagger-spec-validator'
      backend. If False, accepts non-String keys by stringifying them before
      validation. Defaults to True.
    :param str encoding: [optional] For local URLs, use the given file encoding
      instead of auto-detecting. Defaults to None.
    """
    assert url or spec_string and not (url and spec_string), \
        'You must provide either a URL to read, or a spec string to '\
        'parse, but not both!'

    # Keep the parameters around for later use
    self.url = None
    if url:
      from .util.url import absurl
      from .util.fs import abspath
      import os
      self.url = absurl(url, abspath(os.getcwd()))
    else:
      self.url = _PLACEHOLDER_URL

    self._spec_string = spec_string

    # Initialize variables we're filling later
    self.specification = None
    self.version = None
    self.version_name = None
    self.version_parsed = ()
    self.valid = False

    # Add kw args as options
    self.options = kwargs

    # Verify backend
    from .util import default_validation_backend
    self.backend = self.options.get('backend', default_validation_backend())
    if self.backend not in BaseParser.BACKENDS.keys():
      raise ValueError('Backend may only be one of %s!'
              % (BaseParser.BACKENDS.keys(), ))

    # Start parsing if lazy mode is not requested.
    if not lazy:
      self.parse()

  def parse(self):  # noqa: F811
    """
    When the BaseParser was lazily created, load and parse now.

    You can use this function to re-use an existing parser for parsing
    multiple files by setting its url property and then invoking this
    function.
    """
    strict = self.options.get('strict', True)

    # If we have a file name, we need to read that in.
    if self.url and self.url != _PLACEHOLDER_URL:
      from .util.url import fetch_url
      encoding = self.options.get('encoding', None)
      self.specification = fetch_url(self.url, encoding = encoding,
              strict = strict)

    # If we have a spec string, try to parse it.
    if self._spec_string:
      from .util.formats import parse_spec
      self.specification = parse_spec(self._spec_string, self.url)

    # If we have a parsed spec, convert it to JSON. Then we can validate
    # the JSON. At this point, we *require* a parsed specification to exist,
    # so we might as well assert.
    assert self.specification, 'No specification parsed, cannot validate!'

    self._validate()

  def _validate(self):
    # Ensure specification is a mapping
    from collections.abc import Mapping
    if not isinstance(self.specification, Mapping):
      raise ValidationError('Could not parse specifications!')

    # Ensure the selected backend supports the given spec version
    versions, validator_name = BaseParser.BACKENDS[self.backend]

    # Fetch the spec version. Note that this is the spec version the spec
    # *claims* to be; we later set the one we actually could validate as.
    spec_version = None
    if spec_version is None:
      spec_version = self.specification.get('openapi', None)
    if spec_version is None:
      spec_version = self.specification.get('swagger', None)
    if spec_version is None:
      raise ValidationError('Could not determine specification schema '
          'version!')

    # Try parsing the spec version, examine the first component.
    import distutils.version
    parsed = distutils.version.StrictVersion(spec_version).version
    if parsed[0] not in versions:
        raise ValidationError('Version mismatch: selected backend "%s"'
          ' does not support specified version %s!' % (self.backend,
          spec_version))

    # Validate the parsed specs, using the given validation backend.
    validator = getattr(self, validator_name)

    # Set valid flag according to whether validator succeeds
    self.valid = False
    validator(parsed)
    self.valid = True

  def __set_version(self, prefix, version):
    self.version_name = prefix
    self.version_parsed = version
    import semver
    self.semver = str(semver.VersionInfo(*version))

    stringified = self.semver
    if prefix == BaseParser.SPEC_VERSION_2_PREFIX:
      stringified = '%d.%d' % (version[0], version[1])
    self.version = f'{self.version_name} {stringified}'

  def _validate_flex(self, spec_version):  # pragma: nocover
    # Set the version independently of whether validation succeeds
    self.__set_version(BaseParser.SPEC_VERSION_2_PREFIX, spec_version)

    from flex.exceptions import ValidationError as JSEValidationError
    from flex.core import parse as validate
    try:
      validate(self.specification)
    except JSEValidationError as ex:
      from .util.exceptions import raise_from
      raise_from(ValidationError, ex)

  def _validate_swagger_spec_validator(self, spec_version):  # pragma: nocover
    # Set the version independently of whether validation succeeds
    self.__set_version(BaseParser.SPEC_VERSION_2_PREFIX, spec_version)

    from swagger_spec_validator.common import SwaggerValidationError as SSVErr
    from swagger_spec_validator.validator20 import validate_spec
    try:
      validate_spec(self.specification)
    except SSVErr as ex:
      from .util.exceptions import raise_from
      raise_from(ValidationError, ex)

  def _validate_openapi_spec_validator(self, spec_version):  # pragma: nocover
    from openapi_spec_validator import validate_v2_spec, validate_v3_spec
    from jsonschema.exceptions import ValidationError as JSEValidationError
    from jsonschema.exceptions import RefResolutionError

    # Validate according to detected version. Unsupported versions are
    # already caught outside of this function.
    from .util.exceptions import raise_from
    if spec_version[0] == 3:
      # Set the version independently of whether validation succeeds
      self.__set_version(BaseParser.SPEC_VERSION_3_PREFIX, spec_version)

      try:
        validate_v3_spec(self.specification)
      except TypeError as type_ex:
        raise_from(ValidationError, type_ex, self._strict_warning())
      except JSEValidationError as v3_ex:
        raise_from(ValidationError, v3_ex)
      except RefResolutionError as ref_ex:
        raise_from(ValidationError, ref_ex)
    elif spec_version[0] == 2:
      # Set the version independently of whether validation succeeds
      self.__set_version(BaseParser.SPEC_VERSION_2_PREFIX, spec_version)

      try:
        validate_v2_spec(self.specification)
      except TypeError as type_ex:  # pragma: nocover
        raise_from(ValidationError, type_ex, self._strict_warning())
      except JSEValidationError as v2_ex:
        raise_from(ValidationError, v2_ex)
      except RefResolutionError as ref_ex:
        raise_from(ValidationError, ref_ex)

  def _strict_warning(self):
    """Return a warning if strict mode is off."""
    if self.options.get('strict', True):
      return ('Strict mode enabled (the default), so this could be due to an '
          'integer key, such as an HTTP status code.')
    return ('Strict mode disabled. Prance cannot help you narrow this further '
        'down, sorry.')


class ResolvingParser(BaseParser):
  """The ResolvingParser extends BaseParser with resolving references."""

  def __init__(self, url = None, spec_string = None, lazy = False, **kwargs):
    """
    See :py:class:`BaseParser`.

    Resolves JSON pointers/references (i.e. '$ref' keys) before validating the
    specs. The implication is that self.specfication is fully resolved, and
    does not contain any references.

    Additional parameters, see :py::class:`util.RefResolver`.
    """
    # Create a reference cache
    self.__reference_cache = {}

    BaseParser.__init__(
        self,
        url = url,
        spec_string = spec_string,
        lazy = lazy,
        **kwargs
    )

  def _validate(self):
    # We have a problem with the BaseParser's validate function: the
    # jsonschema implementation underlying it does not accept relative
    # path references, but the Swagger specs allow them:
    # http://swagger.io/specification/#referenceObject
    # We therefore use our own resolver first, and validate later.
    from .util.resolver import RefResolver
    forward_arg_names = ('encoding', 'recursion_limit',
            'recursion_limit_handler', 'resolve_types', 'resolve_method',
            'strict')
    forward_args = {
      k: v for (k, v) in self.options.items() if k in forward_arg_names
    }
    resolver = RefResolver(
        self.specification, self.url,
        reference_cache = self.__reference_cache,
        **forward_args
      )
    resolver.resolve_references()
    self.specification = resolver.specs

    # Now validate - the BaseParser knows the specifics
    BaseParser._validate(self)



"""Test suite prance.util.exceptions."""

__author__ = 'Jens Finkhaeuser'
__copyright__ = 'Copyright (c) 2018 Jens Finkhaeuser'
__license__ = 'MIT'
__all__ = ()

import pytest

from prance.util import exceptions
from prance import ValidationError

def test_reraise_without_value_no_extra_message():
  with pytest.raises(ValidationError) as caught:
    exceptions.raise_from(ValidationError, None)

  # The first is obvious from pytest.raises. The rest tests
  # known attributes
  assert caught.type == ValidationError
  assert str(caught.value) == ''


def test_reraise_without_value_extra_message():
  with pytest.raises(ValidationError) as caught:
    exceptions.raise_from(ValidationError, None, 'asdf')

  # The first is obvious from pytest.raises. The rest tests
  # known attributes
  assert caught.type == ValidationError
  assert str(caught.value) == 'asdf'


def test_reraise_with_value_no_extra_message():
  with pytest.raises(ValidationError) as caught:
    try:
      raise RuntimeError("foo")
    except RuntimeError as inner:
      exceptions.raise_from(ValidationError, inner)

  # The first is obvious from pytest.raises. The rest tests
  # known attributes
  assert caught.type == ValidationError
  assert str(caught.value) == 'foo'


def test_reraise_with_value_extra_message():
  with pytest.raises(ValidationError) as caught:
    try:
      raise RuntimeError("foo")
    except RuntimeError as inner:
      exceptions.raise_from(ValidationError, inner, 'asdf')

  # The first is obvious from pytest.raises. The rest tests
  # known attributes
  assert caught.type == ValidationError
  assert str(caught.value) == 'foo -- asdf'


def test_reraise_with_empty_value_string_extra_message():
  with pytest.raises(ValidationError) as caught:
    try:
      raise RuntimeError()
    except RuntimeError as inner:
      exceptions.raise_from(ValidationError, inner, 'asdf')

  # The first is obvious from pytest.raises. The rest tests
  # known attributes
  assert caught.type == ValidationError
  assert str(caught.value) == 'asdf'



"""This module contains code for accessing values in nested data structures."""

__author__ = 'Jens Finkhaeuser'
__copyright__ = 'Copyright (c) 2018 Jens Finkhaeuser'
__license__ = 'MIT'
__all__ = ()


def _json_ref_escape(path):
  """JSON-reference escape object path."""
  path = str(path)  # Could be an int, etc.
  path = path.replace('~', '~0')
  path = path.replace('/', '~1')
  return path


def _str_path(path):
  """Stringify object path."""
  return '/' + '/'.join([_json_ref_escape(p) for p in path])


def path_get(obj, path, defaultvalue = None, path_of_obj = ()):
  """
  Retrieve the value from obj indicated by path.

  Like dict.get(), except:

    - Any Mapping or Sequence is supported.
    - Path is itself a Sequence; the first part is applied to the passed
      object, the second part to the value returned from this operation, and
      so forth recursively.

  :param mixed obj: The Sequence or Mapping from which to retrieve values.
  :param Sequence path: A Sequence of zero or more key/index elements.
  :param mixed defaultvalue: If the value at the path does not exist and this
    parameter is not None, it is returned. Otherwise an error is raised.
  """
  from collections.abc import Mapping, Sequence

  # For error reporting.
  path_of_obj_str = _str_path(path_of_obj)

  if path is not None and not isinstance(path, Sequence):
    raise TypeError('Path is a %s, but must be None or a Collection!'
            % (type(path),))

  if isinstance(obj, Mapping):
    if path is None or len(path) < 1:
      return obj or defaultvalue

    if path[0] not in obj:
      raise KeyError('Object at "{}" does not contain key: {}'.format(
          path_of_obj_str, path[0]))

    return path_get(obj[path[0]], path[1:], defaultvalue,
              path_of_obj = path_of_obj + (path[0],))

  elif isinstance(obj, Sequence):
    if path is None or len(path) < 1:
      return obj or defaultvalue

    try:
      idx = int(path[0])
    except ValueError:
      raise KeyError('Sequence at "%s" needs integer indices only, but got: '
          '%s' % (path_of_obj_str, path[0],))

    if idx < 0 or idx >= len(obj):
      raise IndexError('Index out of bounds for sequence at "%s": %d' % (
          path_of_obj_str, idx))

    return path_get(obj[idx], path[1:], defaultvalue,
           path_of_obj = path_of_obj + (path[0],))

  else:
    # Path must be empty.
    if path is not None and len(path) > 0:
      raise TypeError(f'Cannot get anything from type {type(obj)}!')
    return obj or defaultvalue


def path_set(obj, path, value, **options):
  """
  Set the value in obj indicated by path.

  Setter anologous to path_get() above.

  As setting values is a write operation, this function optionally creates
  intermediate objects to ensure all elements of path can be dereferenced.

  :param mixed obj: The Sequence or Mapping from which to retrieve values.
  :param Sequence path: A Sequence of zero or more key/index elements.
  :param mixed value: The value to set.
  :param bool create: [optional] Flag indicating whether to create
    intermediate values or not. Defaults to False.
  """
  # Retrieve options
  create = options.get('create', False)

  def fill_sequence(seq, index, value_index_type):
    """
    Fill the sequence seq with elements until index can be accessed.

    Fills with None except for the indexed element. That is either a dict or
    a list, depending on the value_index_type. If the latter is an int, a
    list is added. If the latter is None (unknown), None is added. Otherwise
    a dict is added.
    """
    if len(seq) > index:
      return

    while len(seq) < index:
      seq.append(None)

    if value_index_type == int:
      seq.append([])
    elif value_index_type is None:
      seq.append(None)
    else:
      seq.append({})

  def safe_idx(seq, index):
    """
    Safely index a sequence.

    Much like dict.get with default value, except returns None instead of
    raising IndexError.
    """
    try:
      return type(seq[index])
    except IndexError:
      return None

  # print('obj', obj, type(obj))
  # print('path', path)
  # print('value', value)

  from collections.abc import (Sequence, MutableSequence, Mapping,
      MutableMapping)

  if path is not None and not isinstance(path, Sequence):
    raise TypeError('Path is a %s, but must be None or a Collection!'
            % (type(path),))

  if len(path) < 1:
    raise KeyError('Cannot set with an empty path!')

  if isinstance(obj, Mapping):
    # If we don't have a mutable mapping, we should raise a TypeError
    if not isinstance(obj, MutableMapping):  # pragma: nocover
      raise TypeError(f'Mapping is not mutable: {type(obj)}')

    # If the path has only one element, we just overwrite the element at the
    # given key. Otherwise we recurse.
    if len(path) == 1:
      if not create and path[0] not in obj:
        # dicts would normally silently create, but we have to make it
        # explicit to fulfil our contract.
        raise KeyError(f'Key "{path[0]}" not in Mapping!')
      obj[path[0]] = value
    else:
      if create and path[0] not in obj:
        if type(path[1]) == int:
          obj[path[0]] = []
        else:
          obj[path[0]] = {}
      path_set(obj[path[0]], path[1:], value, create = create)

    return obj

  elif isinstance(obj, Sequence):
    idx = path[0]

    # If we don't have a mutable sequence, we should raise a TypeError
    if not isinstance(obj, MutableSequence):
      raise TypeError(f'Sequence is not mutable: {type(obj)}')

    # Ensure integer indices
    try:
      idx = int(idx)
    except ValueError:
      raise KeyError('Sequences need integer indices only.')

    # If we're supposed to create and the index at path[0] doesn't exist,
    # then we need to push some dummy objects.
    if create:
      fill_sequence(obj, idx, safe_idx(path, 1))

    # If the path has only one element, we just overwrite the element at the
    # given index. Otherwise we recurse.
    # print('pl', len(path))
    if len(path) == 1:
      obj[idx] = value
    else:
      path_set(obj[idx], path[1:], value, create = create)

    return obj
  else:
    raise TypeError(f'Cannot set anything on type {type(obj)}!')



"""Test suite for prance.util.path ."""

__author__ = 'Jens Finkhaeuser'
__copyright__ = 'Copyright (c) 2018 Jens Finkhaeuser'
__license__ = 'MIT'
__all__ = ()


import pytest

from prance.util.path import path_set, path_get


def test_get_bad_path():
  # Raise with bad path types
  with pytest.raises(TypeError):
    path_get({}, 42)
  with pytest.raises(TypeError):
    path_get({}, 3.14)

  with pytest.raises(KeyError):
    path_get([], ('a', 'b'))



def test_get_value_default():
  value = 42

  # No path can be resolved in a value type
  result = None
  with pytest.raises(TypeError):
    result = path_get(value, ('foo', 'bar'), 123)
  assert result is None

  # However, we can resolve zero length paths
  result = path_get(value, (), 123)
  assert 42 == result
  result = path_get(None, (), 123)
  assert 123 == result

  # Also we can resolve None-type paths
  result = path_get(value, None, 321)
  assert 42 == result
  result = path_get(None, None, 321)
  assert 321 == result


def test_get_value_no_default():
  value = 42

  # No path can be resolved in a value type
  result = 666
  with pytest.raises(TypeError):
    result = path_get(value, ('foo', 'bar'))
  assert result == 666

  # However, we can resolve zero length paths
  result = path_get(value, ())
  assert 42 == result
  result = path_get(None, ())
  assert result is None

  # Also we can resolve None-type paths
  result = path_get(value, None)
  assert 42 == result
  result = path_get(None, None)
  assert result is None


def test_get_collection_default():
  value = (1, 2, 3)

  # String paths in a Sequence should raise KeyError
  result = None
  with pytest.raises(KeyError):
    result = path_get(value, ('foo', 'bar'), 123)
  assert result is None

  # A numeric path should work, though
  result = path_get(value, (1,), 123)
  assert 2 == result

  # Zero length paths should return the value or default value
  result = path_get(value, (), 123)
  assert (1, 2, 3) == result
  result = path_get(None, (), 123)
  assert 123 == result

  # And None paths as well
  result = path_get(value, None, 321)
  assert (1, 2, 3) == result
  result = path_get(None, None, 321)
  assert 321 == result


def test_get_collection_no_default():
  value = (1, 2, 3)

  # String paths in a Sequence should raise KeyError
  result = None
  with pytest.raises(KeyError):
    result = path_get(value, ('foo', 'bar'))
  assert result is None

  # A numeric path should work, though
  result = path_get(value, (1,))
  assert 2 == result

  # Zero length paths should return the value or default value
  result = path_get(value, ())
  assert (1, 2, 3) == result
  result = path_get(None, ())
  assert result is None

  # And None paths as well
  result = path_get(value, None)
  assert (1, 2, 3) == result
  result = path_get(None, None)
  assert result is None


def test_get_mapping_default():
  value = { 'foo': 1, 'bar': 2, 3: 3 }

  # String paths should work in a Mapping
  result = path_get(value, ('foo',), 123)
  assert 1 == result

  # So should numeric keys
  result = path_get(value, (3,), 123)
  assert 3 == result

  # Zero length paths should return the value or default value
  result = path_get(value, (), 123)
  assert { 'foo': 1, 'bar': 2, 3: 3 } == result
  result = path_get(None, (), 123)
  assert 123 == result

  # And None paths as well
  result = path_get(value, None, 321)
  assert { 'foo': 1, 'bar': 2, 3: 3 } == result
  result = path_get(None, None, 321)
  assert 321 == result



def test_set_bad_path():
  # Raise with bad path types
  with pytest.raises(TypeError):
    path_set({}, 42, None)
  with pytest.raises(TypeError):
    path_set({}, 3.14, None)

  with pytest.raises(KeyError):
    path_set([], ('a', 'b'), None)


def test_set_value():
  value = 42

  # Setting on a value type with a path should raise...
  with pytest.raises(TypeError):
    path_set(42, ('foo', 'bar'), 'something')
  with pytest.raises(TypeError):
    path_set(42, (0, 1), 'something')

  # ... and without a path or an empty path should raise an error
  with pytest.raises(KeyError):
    path_set(42, (), 'something')
  with pytest.raises(TypeError):
    path_set(42, None, 'something')


def test_set_sequence_no_create():
  # Tuples are not mutable - this must raise
  result = None
  with pytest.raises(TypeError):
    result = path_set((42,), (0,), 'something')

  # Lists are mutable - this must work
  result = path_set([42,], (0,), 'something')
  assert ['something'] == result

  # Try nested paths
  result = path_set([42, [1, 2]], (1, 0), 'something')
  assert [42, ['something', 2]] == result

  # Bad paths must raise
  with pytest.raises(IndexError):
    path_set([], (0,), 'something')
  with pytest.raises(IndexError):
    path_set([[]], (0, 1), 'something')


def test_set_sequence_create():
  # If we want to create entries, then mutable sequences should have further
  # sequences added to it.
  result = path_set([], (1, 0), 'something', create = True)
  assert [None, ['something']] == result


def test_set_mapping_no_create():
  # Setting existing keys must work
  result = path_set({ 'foo': 'bar' }, ('foo',), 'new')
  assert { 'foo': 'new' } == result

  # Nested mappings should work just as well.
  result = path_set({ 'foo': { 'bar': 'baz' } }, ('foo', 'bar',), 'new')
  assert { 'foo': { 'bar': 'new' } } == result

  # Bad paths must raise
  with pytest.raises(KeyError):
    path_set({}, ('foo',), 'something')
  with pytest.raises(KeyError):
    path_set({ 'foo': {}}, ('foo', 'bar',), 'something')


def test_set_mapping_create():
  # Setting missing keys must work
  result = path_set({}, ('foo',), 'bar', create = True)
  assert { 'foo': 'bar' } == result

  # Recursive setting must work
  result = path_set({}, ('foo', 'bar',), 'baz', create = True)
  assert { 'foo': { 'bar': 'baz' } } == result


def test_set_mixed_no_create():
  # Sequence in Mapping must work
  result = path_set({ 'foo': [0] }, ('foo', 0), 42)
  assert { 'foo': [42] } == result

  # And Mapping in Sequence likewise
  result = path_set([{ 'foo': 'bar' }], (0, 'foo'), 42)
  assert [{ 'foo': 42 }] == result


def test_set_mixed_create():
  # Creation should work based on the path element types
  result = path_set({}, ('foo', 0), 42, create = True)
  assert { 'foo': [42] } == result

  result = path_set([], (0, 'foo'), 42, create = True)
  assert [{ 'foo': 42 }] == result

  # Create int keys in dict
  result = path_set({}, (0,), 42, create = True)
  assert { 0: 42 } == result


def test_set_mixed_create_no_fill():
  # Creation should not add items that are not necessary
  base = { 'foo': [123] }

  result = path_set(base, ('foo', 0), 42, create = True)
  assert { 'foo': [42] } == result


def test_get_informative_key_error():
  base = { 'foo': { 'bar': [123] } }

  # Match that the object being examing has its path printed, as
  # well as that the key is included.
  with pytest.raises(KeyError, match = r'.*"/".*asdf'):
    path_get(base, ('asdf',))

  with pytest.raises(KeyError, match = r'.*"/foo".*asdf'):
    path_get(base, ('foo', 'asdf'))


def test_get_informative_index_error():
  base = { 'foo': { 'bar': [123] } }

  with pytest.raises(IndexError, match = r'.*"/foo/bar".*123'):
    path_get(base, ('foo', 'bar', 123))



"""Test suite for prance.util.formats ."""

__author__ = 'Jens Finkhaeuser'
__copyright__ = 'Copyright (c) 2016-2021 Jens Finkhaeuser'
__license__ = 'MIT'
__all__ = ()


import pytest

from prance.util import formats


def test_format_info_yaml():
  ctype, ext = formats.format_info('yaml')
  assert 'yaml' in ctype
  assert 'yaml' in ext


def test_format_info_json():
  ctype, ext = formats.format_info('json')
  assert 'json' in ctype
  assert 'json' in ext


def test_format_info_bad_format():
  ctype, ext = formats.format_info('foo')
  assert ctype is None
  assert ext is None


def test_parse_details_yaml():
  yaml = """---
foo: bar
"""
  parsed, ctype, ext = formats.parse_spec_details(yaml, 'foo.yaml')

  assert parsed['foo'] == 'bar', 'Did not parse with explicit YAML'
  assert 'yaml' in ctype
  assert 'yaml' in ext


def test_parse_yaml():
  yaml = """---
foo: bar
"""
  parsed = formats.parse_spec(yaml, 'foo.yaml')
  assert parsed['foo'] == 'bar', 'Did not parse with explicit YAML'

  parsed = formats.parse_spec(yaml)
  assert parsed['foo'] == 'bar', 'Did not parse with implicit YAML'


def test_parse_yaml_ctype():
  yaml = """---
foo: bar
"""
  parsed = formats.parse_spec(yaml, None, content_type = 'text/yaml')
  assert parsed['foo'] == 'bar', 'Did not parse with explicit YAML'


def test_parse_json():
  json = '{ "foo": "bar" }'

  parsed = formats.parse_spec(json, 'foo.js')
  assert parsed['foo'] == 'bar', 'Did not parse with explicit JSON'


def test_parse_json_ctype():
  json = '{ "foo": "bar" }'

  parsed = formats.parse_spec(json, None, content_type = 'application/json')
  assert parsed['foo'] == 'bar', 'Did not parse with explicit JSON'


def test_parse_unknown():
  with pytest.raises(formats.ParseError):
    formats.parse_spec('{-')


def test_parse_unknown_ext():
  with pytest.raises(formats.ParseError):
    formats.parse_spec('{-', 'asdf.xml')


def test_parse_unknown_ctype():
  with pytest.raises(formats.ParseError):
    formats.parse_spec('{-', None, content_type = 'text/xml')


def test_serialize_json():
  specs = {
    'foo': 'bar',
    'baz': [1, 2, 3],
  }

  # With no further information given, the specs must be in JSON
  serialized = formats.serialize_spec(specs)
  assert serialized.startswith('{')

  # The same must be the case if we provide a JSON file name
  serialized = formats.serialize_spec(specs, 'foo.json')
  assert serialized.startswith('{')


def test_serialize_yaml():
  specs = {
    'foo': 'bar',
    'baz': [1, 2, 3],
  }

  # Provide a YAML file name
  serialized = formats.serialize_spec(specs, 'foo.yml')
  assert 'foo: bar' in serialized


def test_serialize_json_ctype():
  specs = {
    'foo': 'bar',
    'baz': [1, 2, 3],
  }

  # Force JSON with content type
  serialized = formats.serialize_spec(specs, None,
                                      content_type = 'application/json')
  assert serialized.startswith('{')


def test_serialize_yaml_ctype():
  specs = {
    'foo': 'bar',
    'baz': [1, 2, 3],
  }

  # Force YAML with content type
  serialized = formats.serialize_spec(specs, None, content_type = 'text/yaml')
  assert 'foo: bar' in serialized



"""Test suite for prance.util ."""

__author__ = 'Jens Finkhaeuser'
__copyright__ = 'Copyright (c) 2016-2021 Jens Finkhaeuser'
__license__ = 'MIT'
__all__ = ()


import pytest

from prance import util


def test_stringify_keys():
  test = {
    'str': 42,
    'unicode': 123,
    'nested': {
      'str': 42,
      'unicode': 123,
    },
    42: 'int',
    3.14: 'float',
  }

  result = util.stringify_keys(test)

  def assert_object_identity(key, first, second):
    ret = None
    for k1, v1 in first.items():
      for k2, v2 in second.items():
        if k1 == key and k2 == key:
          assert id(k1) == id(k2)
          ret = id(k1)
    assert ret is not None
    return ret

  # The number of keys must not have changed
  assert len(test) == len(result)
  assert len(test['nested']) == len(result['nested'])

  # str/unicode keys must remain absolutely unchanged. We're testing for
  # object identity.
  found = []
  for key in ('str', 'unicode'):
    found.append(assert_object_identity(key, test, result))
    found.append(assert_object_identity(key, test['nested'], result['nested']))

  # for all other keys we find in the test set, there must exist a
  # stringified version of it in the result set
  for key, value in test.items():
    if key in found:
      continue

    assert str(key) in result

  for key, value in test['nested'].items():
    if key in found:
      continue

    assert str(key) in result['nested']



"""This submodule contains helpers for exception handling."""

__author__ = 'Jens Finkhaeuser'
__copyright__ = 'Copyright (c) 2018,2019 Jens Finkhaeuser'
__license__ = 'MIT'
__all__ = ()


# Raise the given exception class from the caught exception, preserving
# stack trace and message as much as possible.
def raise_from(klass, from_value, extra_message = None):
  try:
    if from_value is None:
      if extra_message is not None:
        raise klass(extra_message)
      raise klass()

    args = list(from_value.args)
    if extra_message is not None:
      if len(args) and isinstance(args[0], str):
        args[0] += ' -- ' + extra_message
      else:
        args.append(extra_message)
    raise klass(*args) from from_value
  finally:
    klass = None



"""Test suite for prance.cli ."""

__author__ = 'Jens Finkhaeuser'
__copyright__ = 'Copyright (c) 2016-2021 Jens Finkhaeuser'
__license__ = 'MIT'
__all__ = ()

import pytest

from . import none_of

@pytest.fixture
def runner():
  from click.testing import CliRunner
  return CliRunner()


@pytest.mark.skipif(none_of('click'), reason = 'Click does not exist')
def test_validate_defaults(runner):
  from prance import cli

  # Good example
  result = runner.invoke(cli.validate, ['tests/specs/petstore.yaml'])
  assert result.exit_code == 0
  expected = """Processing "tests/specs/petstore.yaml"...
 -> Resolving external references.
Validates OK as Swagger/OpenAPI 2.0!
"""
  assert result.output == expected

  # Bad example
  result = runner.invoke(cli.validate, ['tests/specs/definitions.yaml'])
  assert result.exit_code == 1
  assert 'ValidationError' in result.output


@pytest.mark.skipif(none_of('click'), reason = 'Click does not exist')
def test_validate_multiple(runner):
  from prance import cli

  result = runner.invoke(cli.validate,
      ['tests/specs/petstore.yaml', 'tests/specs/petstore.yaml'])
  assert result.exit_code == 0
  expected = """Processing "tests/specs/petstore.yaml"...
 -> Resolving external references.
Validates OK as Swagger/OpenAPI 2.0!
Processing "tests/specs/petstore.yaml"...
 -> Resolving external references.
Validates OK as Swagger/OpenAPI 2.0!
"""
  assert result.output == expected


@pytest.mark.skipif(none_of('click'), reason = 'Click does not exist')
def test_validate_no_resolve(runner):
  from prance import cli

  # Good example
  result = runner.invoke(cli.validate, ['--no-resolve', 'tests/specs/petstore.yaml'])
  assert result.exit_code == 0
  expected = """Processing "tests/specs/petstore.yaml"...
 -> Not resolving external references.
Validates OK as Swagger/OpenAPI 2.0!
"""
  assert result.output == expected


@pytest.mark.skipif(none_of('click'), reason = 'Click does not exist')
def test_validate_output_too_many_inputs(runner):
  from prance import cli

  result = runner.invoke(cli.validate,
      ['-o', 'foo', 'tests/specs/petstore.yaml', 'tests/specs/petstore.yaml'])
  assert result.exit_code == 2
  assert 'If --output-file is given,' in result.output


@pytest.mark.skipif(none_of('click'), reason = 'Click does not exist')
def test_validate_output(runner):
  from prance import cli

  import os, os.path
  curdir = os.getcwd()

  outnames = ['foo.json', 'foo.yaml']
  for outname in outnames:
    with runner.isolated_filesystem():
      result = runner.invoke(cli.validate,
          ['-o', outname, os.path.join(curdir, 'tests/specs/petstore.yaml')])
      assert result.exit_code == 0

      # There also must be a 'foo' file now.
      files = [f for f in os.listdir('.') if os.path.isfile(f)]
      assert outname in files

      # Ensure a UTF-8 file encoding.
      from prance.util import fs
      assert 'utf-8' in fs.detect_encoding(outname, default_to_utf8 = False,
                                           read_all = True)

      # Now do a full encoding detection, too

      # The 'foo' file must be a valid swagger spec.
      result = runner.invoke(cli.validate, [outname])
      assert result.exit_code == 0


@pytest.mark.skipif(none_of('click'), reason = 'Click does not exist')
def test_compile_defaults(runner):
  from prance import cli

  # Good example
  result = runner.invoke(cli.compile, ['tests/specs/petstore.yaml'])
  assert result.exit_code == 0
  expected = """Processing "tests/specs/petstore.yaml"...
 -> Resolving external references.
Validates OK as Swagger/OpenAPI 2.0!
"""
  assert result.output.startswith(expected)

  # Bad example
  result = runner.invoke(cli.validate, ['tests/specs/definitions.yaml'])
  assert result.exit_code == 1
  assert 'ValidationError' in result.output


@pytest.mark.skipif(none_of('click'), reason = 'Click does not exist')
def test_compile_output(runner):
  from prance import cli

  import os, os.path
  curdir = os.getcwd()

  outnames = ['foo.json', 'foo.yaml']
  for outname in outnames:
    with runner.isolated_filesystem():
      result = runner.invoke(cli.compile,
          [os.path.join(curdir, 'tests/specs/petstore.yaml'), outname])
      assert result.exit_code == 0

      # There also must be a 'foo' file now.
      files = [f for f in os.listdir('.') if os.path.isfile(f)]
      assert outname in files

      # Ensure a UTF-8 file encoding.
      from prance.util import fs
      assert 'utf-8' in fs.detect_encoding(outname, default_to_utf8 = False,
                                           read_all = True)

      # Now do a full encoding detection, too

      # The 'foo' file must be a valid swagger spec.
      result = runner.invoke(cli.validate, [outname])
      assert result.exit_code == 0


@pytest.mark.skipif(none_of('click'), reason = 'Click does not exist')
@pytest.mark.requires_network()
def test_convert_defaults(runner):
  from prance import cli

  # Good example
  result = runner.invoke(cli.convert, ['tests/specs/petstore.yaml'])
  assert result.exit_code == 0
  assert 'openapi' in result.output
  assert '3.' in result.output

  # Bad example
  result = runner.invoke(cli.validate, ['tests/specs/definitions.yaml'])
  assert result.exit_code == 1
  assert 'ValidationError' in result.output


@pytest.mark.skipif(none_of('click'), reason = 'Click does not exist')
@pytest.mark.requires_network()
def test_convert_output(runner):
  from prance import cli

  import os, os.path
  curdir = os.getcwd()

  outnames = ['foo.json', 'foo.yaml']
  for outname in outnames:
    with runner.isolated_filesystem():
      result = runner.invoke(cli.convert,
          [os.path.join(curdir, 'tests/specs/petstore.yaml'), outname])
      assert result.exit_code == 0

      # There also must be a 'foo' file now.
      files = [f for f in os.listdir('.') if os.path.isfile(f)]
      assert outname in files

      # Ensure a UTF-8 file encoding.
      from prance.util import fs
      assert 'utf-8' in fs.detect_encoding(outname, default_to_utf8 = False,
                                           read_all = True)

      # Now do a full encoding detection, too
      contents = fs.read_file(outname)
      assert 'openapi' in contents
      assert '3.' in contents



"""Test suite for prance.ResolvingParser ."""

__author__ = 'Jens Finkhaeuser'
__copyright__ = 'Copyright (c) 2016-2021 Jens Finkhaeuser'
__license__ = 'MIT'
__all__ = ()

import pytest
from unittest.mock import patch

from prance import ResolvingParser
from prance import ValidationError

from . import none_of

def mock_get_petstore(*args, **kwargs):
  from .mock_response import MockResponse, PETSTORE_YAML
  return MockResponse(text = PETSTORE_YAML)


@pytest.fixture
def petstore_parser():
  return ResolvingParser('tests/specs/petstore.yaml')


@pytest.fixture
@patch('requests.get')
def with_externals_parser(mock_get):
  mock_get.side_effect = mock_get_petstore
  return ResolvingParser('tests/specs/with_externals.yaml')


@pytest.fixture
def petstore_parser_from_string():
  yaml = None
  with open('tests/specs/petstore.yaml', 'rb') as f:
    x = f.read()
    yaml = x.decode('utf8')
  return ResolvingParser(spec_string = yaml)


@pytest.fixture
def issue_1_parser():
  return ResolvingParser('tests/specs/issue_1.json')


@pytest.mark.skipif(none_of('openapi-spec-validator', 'swagger-spec-validator', 'flex'), reason='Missing backends')
def test_basics(petstore_parser):
  assert petstore_parser.specification, 'No specs loaded!'


@pytest.mark.skipif(none_of('openapi-spec-validator', 'swagger-spec-validator', 'flex'), reason='Missing backends')
def test_petstore_resolve(petstore_parser):
  assert petstore_parser.specification, 'No specs loaded!'

  # The petstore references /definitions/Pet in /definitions/Pets, and uses
  # /definitions/Pets in the 200 response to the /pets path. So let's check
  # whether we can find something of /definitions/Pet there...
  res = petstore_parser.specification['paths']['/pets']['get']['responses']
  assert res['200']['schema']['type'] == 'array', 'Did not resolve right!'


@pytest.mark.skipif(none_of('openapi-spec-validator', 'swagger-spec-validator', 'flex'), reason='Missing backends')
def test_with_externals_resolve(with_externals_parser):
  assert with_externals_parser.specification, 'No specs loaded!'

  # The specs are a simplified version of the petstore example, with some
  # external references.
  # - Test that the list pets call returns the right thing from the external
  #   definitions.yaml
  res = with_externals_parser.specification['paths']['/pets']['get']
  res = res['responses']
  assert res['200']['schema']['type'] == 'array'

  # - Test that the get single pet call returns the right thing from the
  #   remote petstore definition
  res = with_externals_parser.specification['paths']['/pets/{petId}']['get']
  res = res['responses']
  assert 'id' in res['200']['schema']['required']

  # - Test that error responses contain a message from error.json
  res = with_externals_parser.specification['paths']['/pets']['get']
  res = res['responses']
  assert 'message' in res['default']['schema']['required']


@pytest.mark.skipif(none_of('openapi-spec-validator', 'swagger-spec-validator', 'flex'), reason='Missing backends')
def test_relative_urls_from_string(petstore_parser_from_string):
  # This must succeed
  assert petstore_parser_from_string.yaml(), 'Did not get YAML representation of specs!'


@pytest.mark.skipif(none_of('openapi-spec-validator', 'swagger-spec-validator', 'flex'), reason='Missing backends')
def test_issue_1_relative_path_references(issue_1_parser):
  # Must resolve references correctly
  params = issue_1_parser.specification["paths"]["/test"]["parameters"]
  assert 'id' in params[0]['schema']['required']


@pytest.mark.skipif(none_of('openapi-spec-validator'), reason='Missing backends')
def test_issue_39_sequence_indices():
  # Must not fail to parse
  parser = ResolvingParser('tests/specs/issue_39.yaml', backend = 'openapi-spec-validator')

  # The /useCase path should have two values in its response example.
  example = parser.specification['paths']['/useCase']['get']['responses']['200']['content']['application/json']['examples']['response']
  assert 'value' in example
  assert len(example['value']) == 2

  # However, the /test path should have only one of the strings.
  example = parser.specification['paths']['/test']['get']['responses']['200']['content']['application/json']['example']
  assert example == 'some really long or specific string'


@pytest.mark.skipif(none_of('openapi-spec-validator'), reason='Missing backends')
def test_issue_51_encoding_error():
  # Parsing used to throw - but shouldn't after heuristic change.
  parser = ResolvingParser('tests/specs/issue_51/openapi-main.yaml',
          lazy = True, backend = 'openapi-spec-validator',
          strict = True)

  parser.parse()

  # Parsing with setting an explicit and wrong file encoding should raise
  # an error, effectively reverting to the old behaviour
  parser = ResolvingParser('tests/specs/issue_51/openapi-main.yaml',
          lazy = True, backend = 'openapi-spec-validator',
          strict = True,
          encoding = 'iso-8859-2')

  from yaml.reader import ReaderError
  with pytest.raises(ReaderError):
    parser.parse()



@pytest.mark.skipif(none_of('openapi-spec-validator'), reason='Missing backends')
def test_issue_65_partial_resolution_files():
  specs = '''openapi: "3.0.0"
info:
  title: ''
  version: '1.0.0'
paths: {}
components:
    schemas:
        SampleArray:
            type: array
            items:
              $ref: '#/components/schemas/ItemType'

        ItemType:
          type: integer
'''

  from prance.util import resolver
  parser = ResolvingParser(
    spec_string = specs,
    resolve_types = resolver.RESOLVE_FILES
  )

  from prance.util.path import path_get
  val = path_get(parser.specification, ('components', 'schemas', 'SampleArray', 'items'))
  assert '$ref' in val


@pytest.mark.skipif(none_of('openapi-spec-validator'), reason='Missing backends')
def test_issue_83_skip_propagation():
  # Throw with strict parsing
  parser = ResolvingParser('tests/specs/issue_83/bad_spec.yml',
          lazy = True, backend = 'openapi-spec-validator',
          strict = True)

  with pytest.raises(ValidationError):
    parser.parse()

  # Do not throw with non-strict parsing
  parser = ResolvingParser('tests/specs/issue_83/bad_spec.yml',
          lazy = True, backend = 'openapi-spec-validator',
          strict = False)

  parser.parse()



"""This submodule contains file system utilities for Prance."""

__author__ = 'Jens Finkhaeuser'
__copyright__ = 'Copyright (c) 2016-2019 Jens Finkhaeuser'
__license__ = 'MIT'
__all__ = ()


# Re-define an error for backwards compatibility
FileNotFoundError = FileNotFoundError  # pragma: no cover


# The following constant and function are taken from
# https://stackoverflow.com/questions/9532499/check-whether-a-path-is-valid-in-python-without-creating-a-file-at-the-paths-ta

# Sadly, Python fails to provide the following magic number for us.
_ERROR_INVALID_NAME = 123
"""
Windows-specific error code indicating an invalid pathname.

See Also
----------
https://msdn.microsoft.com/en-us/library/windows/desktop/ms681382%28v=vs.85%29.aspx
    Official listing of all such codes.
"""


# Following Microsoft documentation, set the default read size for detecting
# a file encoding to a multiple of 4k that seems to work well on various OSes
# and volume sizes.
# https://support.microsoft.com/en-us/help/140365/default-cluster-size-for-ntfs-fat-and-exfat
_READ_CHUNK_SIZE = 64 * 1024
"""
Default read size for detecting file encoding.
"""


def is_pathname_valid(pathname):
  """
  Test whether a path name is valid.

  :return: True if the passed pathname is valid on the current OS, False
      otherwise.
  :rtype: bool
  """
  import errno, os
  # If this pathname is either not a string or is but is empty, this pathname
  # is invalid.
  try:
    if not isinstance(pathname, str) or not pathname:
      return False

    # Strip this pathname's Windows-specific drive specifier (e.g., `C:\`)
    # if any. Since Windows prohibits path components from containing `:`
    # characters, failing to strip this `:`-suffixed prefix would
    # erroneously invalidate all valid absolute Windows pathnames.
    _, pathname = os.path.splitdrive(pathname)

    # Directory guaranteed to exist. If the current OS is Windows, this is
    # the drive to which Windows was installed (e.g., the "%HOMEDRIVE%"
    # environment variable); else, the typical root directory.
    import sys
    root_dirname = os.environ.get('HOMEDRIVE', 'C:') \
        if sys.platform == 'win32' else os.path.sep
    assert os.path.isdir(root_dirname)   # ...Murphy and her ironclad Law

    # Append a path separator to this directory if needed.
    root_dirname = root_dirname.rstrip(os.path.sep) + os.path.sep

    # Test whether each path component split from this pathname is valid or
    # not, ignoring non-existent and non-readable path components.
    for pathname_part in pathname.split(os.path.sep):
      try:
        os.lstat(root_dirname + pathname_part)
      except OSError as exc:
        # If an OS-specific exception is raised, its error code
        # indicates whether this pathname is valid or not. Unless this
        # is the case, this exception implies an ignorable kernel or
        # filesystem complaint (e.g., path not found or inaccessible).
        #
        # Only the following exceptions indicate invalid pathnames:
        #
        # * Instances of the Windows-specific "WindowsError" class
        #   defining the "winerror" attribute whose value is
        #   "_ERROR_INVALID_NAME". Under Windows, "winerror" is more
        #   fine-grained and hence useful than the generic "errno"
        #   attribute. When a too-long pathname is passed, for example,
        #   "errno" is "ENOENT" (i.e., no such file or directory) rather
        #   than "ENAMETOOLONG" (i.e., file name too long).
        # * Instances of the cross-platform "OSError" class defining the
        #   generic "errno" attribute whose value is either:
        #   * Under most POSIX-compatible OSes, "ENAMETOOLONG".
        #   * Under some edge-case OSes (e.g., SunOS, *BSD), "ERANGE".
        if hasattr(exc, 'winerror'):  # pragma: nocover
          if exc.winerror == _ERROR_INVALID_NAME:
            return False
        elif exc.errno in {errno.ENAMETOOLONG, errno.ERANGE}:
          return False
  # If a "TypeError" exception was raised, it almost certainly has the
  # error message "embedded NUL character" indicating an invalid pathname.
  except TypeError:  # pragma: nocover
    return False
  # Null-bytes may also cause this, and they are invalid.
  except ValueError:
    return False
  # If no exception was raised, all path components and hence this
  # pathname itself are valid. (Praise be to the curmudgeonly python.)
  else:
    return True
  # If any other exception was raised, this is an unrelated fatal issue
  # (e.g., a bug). Permit this exception to unwind the call stack.
  #
  # Did we mention this should be shipped with Python already?


def from_posix(fname):
  """
  Convert a path from posix-like, to the platform format.

  :param str fname: The filename in posix-like format.
  :return: The filename in the format of the platform.
  :rtype: str
  """
  import sys
  if sys.platform == 'win32':  # pragma: nocover
    if fname[0] == '/':
      fname = fname[1:]
    fname = fname.replace('/', '\\')
  return fname


def to_posix(fname):
  """
  Convert a path to posix-like format.

  :param str fname: The filename to convert to posix format.
  :return: The filename in posix-like format.
  :rtype: str
  """
  import sys
  if sys.platform == 'win32':  # pragma: nocover
    import os.path
    if os.path.isabs(fname):
      fname = '/' + fname
    fname = fname.replace('\\', '/')
  return fname


def abspath(filename, relative_to = None):
  """
  Return the absolute path of a file relative to a reference file.

  If no reference file is given, this function works identical to
  `canonical_filename`.

  :param str filename: The filename to make absolute.
  :param str relative_to: [optional] the reference file name.
  :return: The absolute path
  :rtype: str
  """
  # Create filename relative to the reference, if it exists.
  import os.path
  fname = from_posix(filename)
  if relative_to and not os.path.isabs(fname):
    relative_to = from_posix(relative_to)
    if os.path.isdir(relative_to):
      fname = os.path.join(relative_to, fname)
    else:
      fname = os.path.join(os.path.dirname(relative_to), fname)

  # Make the result canonical
  fname = canonical_filename(fname)
  return to_posix(fname)


def canonical_filename(filename):
  """
  Return the canonical version of a file name.

  The canonical version is defined as the absolute path, and all file system
  links dereferenced.

  :param str filename: The filename to make canonical.
  :return: The canonical filename.
  :rtype: str
  """
  import os, os.path

  path = from_posix(filename)
  while True:
    path = os.path.abspath(path)
    try:
      p = os.path.dirname(path)
      # os.readlink doesn't exist in windows python2.7
      try:
        deref_path = os.readlink(path)
      except AttributeError:  # pragma: no cover
        return path
      path = os.path.join(p, deref_path)
    except OSError:
      return path


def detect_encoding(filename, default_to_utf8 = True, **kwargs):
  """
  Detect the named file's character encoding.

  If the first parts of the file appear to be ASCII, this function returns
  'UTF-8', as that's a safe superset of ASCII. This can be switched off by
  changing the `default_to_utf8` parameter.

  :param str filename: The name of the file to detect the encoding of.
  :param bool default_to_utf8: Defaults to True. Set to False to disable
      treating ASCII files as UTF-8.
  :param bool read_all: Keyword argument; if True, reads the entire file
      for encoding detection.
  :return: The file encoding.
  :rtype: str
  """
  # Read some of the file
  import os.path
  filename = from_posix(filename)
  file_len = os.path.getsize(filename)
  read_len = min(_READ_CHUNK_SIZE, file_len)

  # ... unless we're supposed to!
  if kwargs.get('read_all', False):
    read_len = file_len

  # Read the first read_len bytes raw, so we can detect the encoding
  with open(filename, 'rb') as raw_handle:
    raw = raw_handle.read(read_len)

  # Detect the encoding the file specfies, if any.
  import codecs
  if raw.startswith(codecs.BOM_UTF8):
    encoding = 'utf-8-sig'
  else:
    # Detect encoding using the best detector available
    try:
      # First try ICU. ICU will report ASCII in the first 32 Bytes as
      # ISO-8859-1, which isn't exactly wrong, but maybe optimistic.
      import icu
      encoding = icu.CharsetDetector(raw).detect().getName().lower()
    except ImportError:  # pragma: nocover
      # If that doesn't work, try chardet - it's not got native components,
      # which is a bonus in some environments, but it's not as precise.
      import chardet
      encoding = chardet.detect(raw)['encoding'].lower()

      # Chardet is more brutal in that it reports ASCII if none of the first
      # Bytes contain high bits. To emulate ICU, we just bump up the detected
      # encoding.
      if encoding == 'ascii':
        encoding = 'iso-8859-1'

    # Both chardet and ICU may detect ISO-8859-x, which may not be possible
    # to decode as UTF-8. So whatever they report, we'll try decoding as
    # UTF-8 before reporting it.
    if default_to_utf8 and encoding in ('ascii', 'iso-8859-1', 'windows-1252'):
      # Try decoding as utf-8
      try:
        raw.decode('utf-8')
        # If this worked... well there's no guarantee it's utf-8, to be
        # honest.
        encoding = 'utf-8'
      except UnicodeDecodeError:
        # Decoding as utf-8 failed, so we can't default to it.
        pass

  return encoding


def read_file(filename, encoding = None):
  """
  Read and decode a file, taking BOMs into account.

  :param str filename: The name of the file to read.
  :param str encoding: The encoding to use. If not given, detect_encoding is
      used to determine the encoding.
  :return: The file contents.
  :rtype: unicode string
  """
  filename = from_posix(filename)
  if not encoding:
    # Detect encoding
    encoding = detect_encoding(filename)

  # Finally, read the file in the detected encoding
  with open(filename, encoding = encoding) as handle:
    return handle.read()


def write_file(filename, contents, encoding = None):
  """
  Write a file with the given encoding.

  The default encoding is 'utf-8'. It's recommended not to change that for
  JSON or YAML output.

  :param str filename: The name of the file to read.
  :param str contents: The file contents to write.
  :param str encoding: The encoding to use. If not given, detect_encoding is
      used to determine the encoding.
  """
  if not encoding:
    encoding = 'utf-8'

  fname = from_posix(filename)
  with open(fname, mode = 'w', encoding = encoding) as handle:
    handle.write(contents)



"""This submodule contains utility code for Prance."""

__author__ = 'Jens Finkhaeuser'
__copyright__ = 'Copyright (c) 2016-2021 Jens Finkhaeuser'
__license__ = 'MIT'
__all__ = ('iterators', 'fs', 'formats', 'resolver', 'url', 'path',
    'exceptions')


def stringify_keys(data):
  """
  Recursively stringify keys in a dict-like object.

  :param dict-like data: A dict-like object to stringify keys in.
  :return: A new dict-like object of the same type with stringified keys,
      but the same values.
  """
  from collections.abc import Mapping
  assert isinstance(data, Mapping)

  ret = type(data)()
  for key, value in data.items():
    if not isinstance(key, str):
      key = str(key)
    if isinstance(value, Mapping):
      value = stringify_keys(value)
    ret[key] = value
  return ret


def validation_backends():
  """Return a list of validation backends supported by the environment."""
  ret = []

  try:
    import flex  # noqa: F401
    ret.append('flex')  # pragma: nocover
  except (ImportError, SyntaxError):  # pragma: nocover
    pass

  try:
    import openapi_spec_validator  # noqa: F401
    ret.append('openapi-spec-validator')  # pragma: nocover
  except (ImportError, SyntaxError):  # pragma: nocover
    pass

  try:
    import swagger_spec_validator  # noqa: F401
    ret.append('swagger-spec-validator')  # pragma: nocover
  except (ImportError, SyntaxError):  # pragma: nocover
    pass

  return tuple(ret)


def default_validation_backend():
  """Return the default validation backend, or raise an error."""
  backends = validation_backends()
  if len(backends) <= 0:  # pragma: nocover
    raise RuntimeError('No validation backend available! Install one of '
        '"flex", "openapi-spec-validator" or "swagger-spec-validator".')
  return backends[0]



"""Test suite for prance.convert ."""

__author__ = 'Jens Finkhaeuser'
__copyright__ = 'Copyright (c) 2018-2021 Jens Finkhaeuser'
__license__ = 'MIT'
__all__ = ()

import pytest

from . import none_of

from prance import convert

@pytest.fixture
def petstore_yaml():
  from prance.util import fs
  return fs.read_file('tests/OpenAPI-Specification/examples/v2.0/yaml/petstore.yaml')

@pytest.fixture
def petstore_json():
  from prance.util import fs
  return fs.read_file('tests/OpenAPI-Specification/examples/v2.0/json/petstore.json')


@pytest.mark.requires_network()
def test_convert_petstore_yaml(petstore_yaml):
  converted, content_type = convert.convert_str(petstore_yaml)

  # Check correct content type
  assert 'yaml' in content_type

  # Parsing can't fail.
  from prance.util import formats
  parsed = formats.parse_spec(converted, content_type = content_type)

  # Assert the correct target version
  assert 'openapi' in parsed
  assert parsed['openapi'].startswith('3.')


@pytest.mark.requires_network()
def test_convert_petstore_json(petstore_json):
  converted, content_type = convert.convert_str(petstore_json)

  # Check correct content type
  assert 'json' in content_type

  # Parsing can't fail.
  from prance.util import formats
  parsed = formats.parse_spec(converted, content_type = content_type)

  # Assert the correct target version
  assert 'openapi' in parsed
  assert parsed['openapi'].startswith('3.')


@pytest.mark.requires_network()
def test_convert_petstore_yaml_explicit_name(petstore_yaml):
  converted, content_type = convert.convert_str(petstore_yaml, filename = 'foo.yml')

  # Check correct content type
  assert 'yaml' in content_type


@pytest.mark.requires_network()
def test_convert_url():
  from prance.util import url
  converted, content_type = convert.convert_url(url.absurl('python://tests/specs/petstore.yaml'))

  # Check correct content type
  assert 'yaml' in content_type

  # Parsing can't fail.
  from prance.util import formats
  parsed = formats.parse_spec(converted, content_type = content_type)

  # Assert the correct target version
  assert 'openapi' in parsed
  assert parsed['openapi'].startswith('3.')


@pytest.mark.requires_network()
@pytest.mark.xfail()
def test_convert_spec():
  from prance import BaseParser, ResolvingParser, ValidationError
  parser = BaseParser('tests/specs/petstore.yaml')

  # Conversion should fail with the default backend.
  with pytest.raises(ValidationError):
    converted = convert.convert_spec(parser.specification)

  # However, with the lazy flag it should work.
  converted = convert.convert_spec(parser.specification, lazy = True)
  assert isinstance(converted, BaseParser)

  # Passing a ResolvingParser class should also work.
  converted = convert.convert_spec(parser.specification, ResolvingParser, lazy = True)
  assert isinstance(converted, ResolvingParser)


@pytest.mark.requires_network()
@pytest.mark.xfail()
def test_convert_parser_lazy_swagger_backend():
  from prance import BaseParser, ResolvingParser, ValidationError
  parser = BaseParser('tests/specs/petstore.yaml')

  # Conversion should fail with the default backend.
  with pytest.raises(ValidationError):
    converted = convert.convert_spec(parser)

  # However, with the lazy flag it should work.
  converted = convert.convert_spec(parser, lazy = True)
  assert isinstance(converted, BaseParser)

  # Passing a ResolvingParser class should also work.
  converted = convert.convert_spec(parser, ResolvingParser, lazy = True)
  assert isinstance(converted, ResolvingParser)


@pytest.mark.skipif(none_of('openapi-spec-validator'), reason='Missing openapi-spec-validator')
@pytest.mark.requires_network()
def test_convert_parser_validated():
  from prance import BaseParser
  parser = BaseParser('tests/specs/petstore.yaml', backend = 'openapi-spec-validator')

  # Conversion should work: it's the right backend, and it validates.
  converted = convert.convert_spec(parser)
  assert isinstance(converted, BaseParser)
  assert converted.version_parsed[0] == 3



"""
Defines Mixins for parsers.

The Mixins are here mostly for separation of concerns.
"""

__author__ = 'Jens Finkhaeuser'
__copyright__ = 'Copyright (c) 2016-2018 Jens Finkhaeuser'
__license__ = 'MIT'
__all__ = ()


class CacheSpecsMixin:
  """
  CacheSpecsMixin helps determine if self.specification changed.

  It does so by caching a shallow copy on-demand.
  """

  __CACHED_SPECS = '__cached_specs'

  def specs_updated(self):
    """
    Test if self.specficiation changed.

    :return: Whether the specs changed.
    :rtype: bool
    """
    # Cache specs and return true if no specs have been cached
    if not getattr(self, self.__CACHED_SPECS, None):
      setattr(self, self.__CACHED_SPECS, self.specification.copy())
      return True

    # If specs have been cached, compare them to the current
    # specs.
    cached = getattr(self, self.__CACHED_SPECS)
    if cached != self.specification:
      setattr(self, self.__CACHED_SPECS, self.specification.copy())
      return True

    # Return false if they're the same
    return False


class YAMLMixin(CacheSpecsMixin):
  """
  YAMLMixin returns a YAML representation of the specification.

  It uses :py:class:`CacheSpecsMixin` for lazy evaluation.
  """

  __YAML = '__yaml'

  def yaml(self):
    """
    Return a YAML representation of the specifications.

    :return: YAML representation.
    :rtype: dict
    """
    # Query specs_updated first to start caching
    if self.specs_updated() or not getattr(self, self.__YAML, None):
      import yaml
      setattr(self, self.__YAML, yaml.dump(self.specification))
    return getattr(self, self.__YAML)


class JSONMixin(CacheSpecsMixin):
  """
  JSONMixin returns a JSON representation of the specification.

  It uses :py:class:`CacheSpecsMixin` for lazy evaluation.
  """

  __JSON = '__json'

  def json(self):
    """
    Return a JSON representation of the specifications.

    :return: JSON representation.
    :rtype: dict
    """
    # Query specs_updated first to start caching
    if self.specs_updated() or not getattr(self, self.__JSON, None):
      import json
      setattr(self, self.__JSON, json.dumps(self.specification))
    return getattr(self, self.__JSON)



"""Utility code for tests."""

__author__ = 'Jens Finkhaeuser'
__copyright__ = 'Copyright (c) 2017-2021 Jens Finkhaeuser'
__license__ = 'MIT'
__all__ = ()

import pytest

class Sandboxer:
  """
  Given a sandbox template path, uses pytest to create a temporary directory,
  then copies the template's contents there, and changes to the directory.

  At exit, it changes back to the current working directory.
  """
  def __init__(self, tmpdir, path = None):
    self.__tmpdir = tmpdir
    self.__templatepath = path


  def __enter__(self):
    # Save current working directory & python path
    import os, sys
    self.__cwd = os.getcwd()
    self.__syspath = sys.path[:]

    # Create a new temporary directory and copy the template over
    if self.__templatepath:
      tmp = str(self.__tmpdir.join(self.__templatepath))
      import shutil
      shutil.copytree(self.__templatepath, tmp)
    else:
      tmp = self.__tmpdir

    # Then change to the temporary directory & python path
    os.chdir(str(tmp))
    import site
    sys.path.insert(0, tmp)


  def __exit__(self, exc_type, exc_value, traceback):
    # Restore current working directory & python path
    import os, sys
    os.chdir(str(self.__cwd))
    sys.path = self.__syspath

    # Don't suppress exceptions
    return False


def sandbox(*args, **kwargs):
  """Create a sandbox"""
  return Sandboxer(*args, **kwargs)



"""Test suite for prance.util.fs ."""

__author__ = 'Jens Finkhaeuser'
__copyright__ = 'Copyright (c) 2016-2021 Jens Finkhaeuser'
__license__ = 'MIT'
__all__ = ()


import os
import sys

import pytest

from prance.util import fs

from . import sandbox, platform

@pytest.fixture
def create_symlink():
  # Defined as a fixture so we can selectively use it for tests."
  testname = 'tests/specs/symlink_test'
  import os, os.path
  if os.path.islink(testname):
    return testname

  if os.path.exists(testname):
    os.unlink(testname)

  target = 'with_externals.yaml'
  os.symlink(target, testname)

  return testname


@pytest.mark.skipif(platform('win32'), reason = 'Skip on win32')
def test_canonical(create_symlink):
  res = fs.canonical_filename(create_symlink)
  expected = os.path.join(os.getcwd(), 'tests/specs/with_externals.yaml')
  assert res == expected


def test_to_posix_rel():
  test = "tests/specs/with_externals.yaml"
  assert fs.to_posix(os.path.normpath(test)) == test


@pytest.mark.skipif(platform('win32'), reason = 'Skip on win32')
def test_to_posix_abs_posix():
  test = "/etc/passwd"
  expected = test
  assert fs.to_posix(test) == expected


@pytest.mark.skipif(platform('!win32'), reason = 'Skip on !win32')
def test_to_posix_abs_win32():
  test = "c:\\windows\\notepad.exe"
  expected = "/c:/windows/notepad.exe"
  assert fs.to_posix(test) == expected


def test_from_posix_rel():
  test = "tests/specs/with_externals.yaml"
  assert fs.from_posix(test) == os.path.normpath(test)


@pytest.mark.skipif(platform('win32'), reason = 'Skip on win32')
def test_from_posix_abs_posix():
  test = "/etc/passwd"
  expected = test
  assert fs.from_posix(test) == expected


@pytest.mark.skipif(platform('!win32'), reason = 'Skip on !win32')
def test_from_posix_abs_win32():
  test = "/c:/windows/notepad.exe"
  expected = "c:\\windows\\notepad.exe"
  assert fs.from_posix(test) == expected


def test_abspath_basics():
  testname = os.path.normpath('tests/specs/with_externals.yaml')
  res = fs.abspath(testname)
  expected = fs.to_posix(os.path.join(os.getcwd(), testname))
  assert res == expected


def test_abspath_relative():
  testname = 'error.json'
  relative = os.path.join(os.getcwd(), 'tests/specs/with_externals.yaml')
  res = fs.abspath(testname, relative)
  expected = fs.to_posix(os.path.join(os.getcwd(), 'tests', 'specs', testname))
  assert res == expected


def test_abspath_relative_dir():
  testname = 'error.json'
  relative = os.path.join(os.getcwd(), 'tests', 'specs')
  res = fs.abspath(testname, relative)
  expected = fs.to_posix(os.path.join(os.getcwd(), 'tests', 'specs', testname))
  assert res == expected


def test_detect_encoding():
  # Quick detection should yield utf-8 for the petstore file.
  assert fs.detect_encoding('tests/specs/petstore.yaml') == 'utf-8'

  # Without defaulting to utf-8, it should also work.
  assert fs.detect_encoding('tests/specs/petstore.yaml',
                            default_to_utf8 = False) == 'utf-8'

  # Deep inspection should not change anything because the file size is
  # too small.
  assert fs.detect_encoding('tests/specs/petstore.yaml',
                            default_to_utf8 = False,
                            read_all = True) == 'utf-8'

  # The UTF-8 file with BOM should be detected properly
  assert fs.detect_encoding('tests/specs/utf8bom.yaml') == 'utf-8-sig'


def test_issue_51_detect_encoding():
  # This should be UTF-8, but in issue 51 it was reported that it comes back
  # as iso-8859-2

  # Detection should be ok if the entire file is read
  assert fs.detect_encoding('tests/specs/issue_51/openapi-part.yaml', read_all = True) == 'utf-8'

  # After the heuristic change, it reads the whole file anyway.
  assert fs.detect_encoding('tests/specs/issue_51/openapi-part.yaml') == 'utf-8'

  # Specifically re-encoded as iso-8859-2 should fail - but not as
  # a call to the detect_encoding() function. Instead, we can only return
  # a badly detected encoding. Chardet sends iso-8859-1 here.
  assert fs.detect_encoding('tests/specs/issue_51/openapi-part-iso-8859-2.yaml') == 'iso-8859-1'


def test_load_nobom():
  contents = fs.read_file('tests/specs/petstore.yaml')
  assert contents.index('Swagger Petstore') >= 0, 'File reading failed!'


def test_load_utf8bom():
  contents = fs.read_file('tests/specs/utf8bom.yaml')
  assert contents.index('söme välüe') >= 0, 'UTF-8 BOM handling failed!'


def test_load_utf8bom_override():
  with pytest.raises(UnicodeDecodeError):
    fs.read_file('tests/specs/utf8bom.yaml', 'ascii')


def test_write_file(tmpdir):
  with sandbox.sandbox(tmpdir):
    test_text = 'söme täxt'
    fs.write_file('test.out', test_text)

    # File must have been written
    files = [f for f in os.listdir('.') if os.path.isfile(f)]
    assert 'test.out' in files

    # File contents must work
    contents = fs.read_file('test.out')
    assert test_text == contents


def test_write_file_bom(tmpdir):
  with sandbox.sandbox(tmpdir):
    test_text = 'söme täxt'
    fs.write_file('test.out', test_text, 'utf-8-sig')

    # File must have been written
    files = [f for f in os.listdir('.') if os.path.isfile(f)]
    assert 'test.out' in files

    # Encoding must match the one we've given
    encoding = fs.detect_encoding('test.out')
    assert encoding == 'utf-8-sig'

    # File contents must work
    contents = fs.read_file('test.out')
    assert test_text == contents


def test_valid_pathname():
  # A URL should not be valid
  from prance.util.fs import is_pathname_valid
  assert False == is_pathname_valid('\x00o.bar.org')

  # However, the current path should be.
  import os
  assert True == is_pathname_valid(os.getcwd())

  # Can't put non-strings into this function
  assert True == is_pathname_valid('foo')
  assert False == is_pathname_valid(123)

  # Can't accept too long components
  assert False == is_pathname_valid('a'*256)
